#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 10:59:10 2020

@author: villa
"""

from pymatgen.io.vasp.inputs import Kpoints, Poscar, Potcar, VaspInput, Incar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath

from pynter import SETTINGS


class DefaultInputs:
    """
    Class to generate VASP input files with default parameters. Use with extreme care.    
    """
    
    def __init__(self,structure=None,sort_structure=True):
        """
        Parameters
        ----------
        structure : Pymatgen Structure object. The default is None. If no Structure is given only default INCAR can be generated.
        sort_structure (bool) : Sort structure to avoid POSCAR/POTCAR inconsistencies. The default is True.
        """
        if structure and sort_structure:
            structure.sort()
        self._structure = structure if structure else None 
        defaults = SETTINGS['vasp']
        
        self.default_potcar_symbols =  defaults['default_potcar_symbols']
        if self.structure:
            self.potcar_symbols = [self.default_potcar_symbols[el.symbol] for el in self.structure.composition.elements]
        self.incar_default_flags = defaults['incar_default']
        
    @property
    def structure(self):
        return self._structure
    
    @structure.setter
    def structure(self,structure):
        self._structure = structure

    def get_incar_default(self, xc='PBE', ldauu=None, aexx=0.25):
        """
        Get dictionary with default INCAR parameters

        Parameters
        ----------
        xc : (str), optional
            Which functional to use('PBE','PBE+U','LDA','LDA+U',HSE06' available). The default is 'PBE'.
        ldauu : (List), optional
            List of integers for U parameter to be given in order as required by VASP. The default is None. If None no 'LDAUU' flag is written.
        aexx : (Int), optional
            Value of Hartree-Fock contribution for HSE calculation. The default is 0.25.

        Returns
        -------
        incar_default_flags : Dict
            Dictionary with default INCAR parameters.

        """
        system = self.structure.composition.reduced_formula if self.structure else 'No system info'
          
        incar_default_flags = self.incar_default_flags              
        incar_default_flags["SYSTEM"] = system                   
                     
        if xc == 'PBE' or xc == 'LDA':
            incar_default_flags.update({
                    "ISYM":2})
            
        if xc == 'PBE+U' or xc == 'LDA+U':
            incar_default_flags.update({
                    "ISYM":2,
                    "LDAU" : ".TRUE.",
                    "LDAUTYPE": 2,
                    "LDAUPRINT": 2,
                })
            if ldauu:
                incar_default_flags.update({"LDAUU": ' '.join([str(u) for u in ldauu])}) # if LDAUU is not present VASP will put all zeros
            
        if xc == 'HSE06':
            incar_default_flags.update({
                    "LHFCALC" : ".TRUE.",
                    "HFSCREEN": 0.2,
                    "PRECFOCK": "Fast",
                    "AEXX": aexx,
                    "ISYM":3
                    })
    
        return incar_default_flags


    def get_vasp_input(self, xc='PBE', ldauu=None, aexx=0.25, kppa=1000,
                       potcar_symbols=None, potcar_functional='PBE'):
        """
        Get pymatgen VaspInput object with complete set of default inputs.

        Parameters
        ----------
        xc : (str), optional
            Which functional to use('PBE','PBE+U','HSE06' available). The default is 'PBE'.
        ldauu : (List), optional
            List of integers for U parameter to be given in order as required by VASP. The default is None. If None no 'LDAUU' flag is written.
        aexx : (Int), optional
            Value of Hartree-Fock contribution for HSE calculation. The default is 0.25.
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.
        potcar_symbols : List, optional
            List of strings with symbols of POTCARs. The default is None. If None the default symbols from the Materials Project \n
            database are used for every specie in the given Structure object.
        potcar_functional : (str), optional
            Functional to be used, as defined by Pymatgen Potcar class. The default is 'PBE'.

        Returns
        -------
        Pymatgen VaspInput object.

        """
        
        incar = Incar(self.get_incar_default(xc=xc,ldauu=ldauu,aexx=aexx))
        kpoints = self.get_kpoints_default(kppa=kppa)
        poscar = self.get_poscar()
        potcar = self.get_potcar(potcar_symbols=potcar_symbols,potcar_functional=potcar_functional)
        
        return VaspInput(incar,kpoints,poscar,potcar)
        
        
    def get_kpoints_default(self,kppa=1000):
        """
        Get Pymatgen Kpoints object with automatic Gamma centered k-mesh generation from given Structure object.

        Parameters
        ----------
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.

        Returns
        -------
        Pymatgen Kpoint object.
        """
        
        if self.structure:
            structure = self.structure
        else:
            raise ValueError('Structure object needs to be given to obtain pymatgen automatic kpoint mesh')
         
        return Kpoints().automatic_gamma_density(structure,kppa)
    
  
    def get_kpoints_bs_default(self,divisions=10,hybrid_mode=False,kppa=1000):
        """
        Get Pymatgen Kpoints object for band structure calculations.
        The default high symmetry path for PBE from Pymatgen class HighSymmKpath is obtained from 
        the input Structure with 10 points between high symm k-points.
        For Hybrid BS the calculations need to be selfconsistent, the KPOINTS file is a list of single kpoints.
        The irriducible kpoints with weight != 0 are obtained with Pymatgen's SpaceGroupAnalyzer.
        The additional points with weight = 0 on the high symm path are obtained with the HighSymmPath class
        like for the PBE case.'

        Parameters
        ----------
        divisions : int, optional
            Number of points in between high symmetry points. The default is 10.
        hybrid_mode : (bool), optional
            Kpoints for hybrid calculation. The default is False.
        kppa : (Int), optional
            Value of k-points density per atom to generate automatic k-mesh as done by Pymatgen. The default is 1000.

        Returns
        -------
        Pymatgen Kpoint object.
        """
        if hybrid_mode: 
            # code copy/pasted (slightly adjusted) from pymatgen.io.vasp.sets.MPHSEBSSet
            # The whole input set was not needed but only the kpoints generation in the kpoints @property
            kpts = []
            weights = []  
            all_labels = []  
            structure = self.structure
    
            # for both modes, include the Uniform mesh w/standard weights
            grid = Kpoints().automatic_gamma_density(structure, kppa).kpts
            ir_kpts = SpacegroupAnalyzer(structure, symprec=0.1).get_ir_reciprocal_mesh(grid[0])
            for k in ir_kpts:
                kpts.append(k[0])
                weights.append(int(k[1]))
                all_labels.append('')
       
            #add the symmetry lines w/zero weight
            kpath = HighSymmKpath(structure)
            frac_k_points, labels = kpath.get_kpoints(line_density=divisions,coords_are_cartesian=False)

            for k, f in enumerate(frac_k_points):
                kpts.append(f)
                weights.append(0.0)
                all_labels.append(labels[k])
    
            comment = "HSE run along symmetry lines"   
            kpoints =  Kpoints(comment=comment,style=Kpoints.supported_modes.Reciprocal,num_kpts=len(kpts),
                kpts=kpts,kpts_weights=weights,labels=all_labels)

        else:
            kpoints = Kpoints().automatic_linemode(10,HighSymmKpath(self.structure))
    
        return kpoints
    
    
    def get_poscar(self):
        """
        Get Pymatgen Poscar object from given Structure object.        

        Returns
        -------
        Pymatgen Poscar object
        """
        
        if self.structure:
            pass
        else:
            raise ValueError('Structure object needs to be given to get Poscar onject')
        
        return Poscar(self.structure)
        
     
    def get_potcar(self,potcar_symbols=None,potcar_functional='PBE'):
        """
        Get Pymatgen Potcar object. The PMG_VASP_PSP directory has to be defined in .pmgrc.yaml file as required by Pymatgen.
        Parameters
        ----------
        potcar_symbols : List, optional
            List of strings with symbols of POTCARs. The default is None. If None the default symbols from the Materials Project \n
            database are used for every specie in the given Structure object.
        potcar_functional : (str), optional
            Functional to be used, as defined by Pymatgen Potcar class. The default is 'PBE'.

        Returns
        -------
        Pymatgen Potcar object.
        """
                
        if potcar_symbols:
            potcar = Potcar(symbols=potcar_symbols,functional=potcar_functional)
            self.potcar_symbols = potcar_symbols
        else:
            if self.structure:
                potcar_symbols = self.potcar_symbols
            else:
                raise ValueError('potcar symbols or Structure object need to be given to get Potcar onject')
            potcar = Potcar(symbols=potcar_symbols,functional=potcar_functional)
        
        return potcar
            



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 18 13:12:05 2020

@author: lorenzo
"""

import numpy as np
from pymatgen.core.periodic_table import Element
from pynter.phase_diagram.chempots import Chempots, PDHandler, PressureReservoirs

class OxygenPressure:
    
    def __init__(self,temperature=300,partial_pressure=1):
        """
        Class to extract and organize data regarding chemical potentials as function of oxygen partial pressure.
        Useful for defects analysis for oxides.

        Parameters
        ----------
        temperature : (float), optional
            Temperature. The default is 300.
        partial_pressure : (float), optional
            Partial pressure (p/p0). The default is 1.
        """
        self.temperature = temperature
        self.partial_pressure = partial_pressure


    def chempot_ideal_gas(self, mu0, temperature=None,partial_pressure=None):
        """
        Get chemical potential at a given temperature and partial pressure. The chemical potential in standard conditions (mu0)
        has to be know.

        Parameters
        ----------
        mu0 : (float)
            Value of chemical potential in standard conditions.
        temperature : (float), optional
            Temperature. The default is None. If None the value initialized with the class is used.
        partial_pressure : (float), optional
            Partial pressure. The default is None. If None the value initialized with the class is used.

        Returns
        -------
        chempot : (float)
            Value of chemical potential at given T and p/p0.
        """
        temperature = temperature if temperature else self.temperature
        partial_pressure = partial_pressure if partial_pressure else self.partial_pressure
        kb = 8.6173324e-5  # eV / K
        chempot = mu0 + 0.5*kb*temperature*np.log(partial_pressure)
        return chempot
        

    def get_oxygen_chempot_from_pO2(self,temperature=None,partial_pressure=None):
        """
        Get oxygen chemical potential (delta) from temperature and partial pressure

        Parameters
        ----------
        temperature : (float), optional
            Temperature. The default is None. If None the value initialized with the class is used.
        partial_pressure : (float), optional
            Partial pressure. The default is None. If None the value initialized with the class is used.

        Returns
        -------
        (float)
            Value of oxygen chemical potential (delta) at given T and p/p0.
        """
        temperature = temperature if temperature else self.temperature
        partial_pressure = partial_pressure if partial_pressure else self.partial_pressure
        muO = self.oxygen_standard_chempot(temperature)
        return self.chempot_ideal_gas(muO,temperature,partial_pressure)        
    
    
    def get_pressure_reservoirs_from_pd(self,phase_diagram,target_comp,temperature=None,
                                        extrinsic_chempots_range=None,pressure_range=(-20,10),npoints=50,
                                        get_pressures_as_strings=False):
        """
        Generate Reservoirs object with a set of different chemical potentials starting from a range of oxygen partial pressure.
        The code distinguishes between 2-component and 3-component phase diagrams.
        
        In the case of a 2-comp PD the other chemical potential is calculated directly from the formation energy of the phase
        with target composition.
        In the case of a 3-comp PD the set of chemical potentials is obtained first calculating the boundary phases starting
        from the constant value of muO, then the arithmetic average between this two points in the stability diagram is taken.
        
        In the case where there are some extrinsic elements not belonging to the PD, they must be added in the 
        extrinsic_chempots_range dictionary ({Element:(O_poor_chempot,O_rich_chempot)}). The chempots of extrinsic elements 
        are interpolated linearly between O_poor and O_rich values. 
        
        Parameters
        ----------
        phase_diagram : (PhaseDiagram)
            Pymatgen PhaseDiagram object.
        target_comp : (Composition)
            Pymatgen Composition object.
        temperature : (float), optional
            Temperature value. If None the value with which the class is initialized is taken. The default is None.
        extrinsic_chempots_range : (dict)
            Dictionary with chemical potentials of elements not belonging to the PD ({Element:(O_poor_chempot,O_rich_chempot)}). 
            The default is None.
        pressure_range : (tuple), optional
            Exponential range in which to evaluate the partial pressure . The default is from 1e-20 to 1e10.
        npoints : (int), optional
            Number of data points to interpolate the partial pressure with. The default is 50.
        get_pressures_as_strings : (bool), optional
            Get pressure values (keys in the Reservoirs dict) as strings. The default is set to floats.

        Returns
        -------
        reservoirs
            PressureReservoirs object. The dictionary is organized as {partial_pressure:chempots}.
        """
        chempots_dict = {}
        reservoirs = {}
        pd = phase_diagram
        temperature = temperature if temperature else self.temperature
        partial_pressures = np.logspace(pressure_range[0],pressure_range[1],num=npoints,base=10)
        mu_standard = self.oxygen_standard_chempot(temperature)
        
        i = 0
        for p in partial_pressures:
            i += 1
            muO = self.chempot_ideal_gas(mu_standard,temperature=temperature,partial_pressure=p) #delta value
            fixed_chempot = Chempots({'O':muO})
            
            if len(pd.elements) == 2: # 2-component PD case
                pdh = PDHandler(pd)
                mu = pdh.calculate_single_chempot(target_comp, fixed_chempot)
                for el in target_comp.elements:
                    if el is not Element('O'):
                        chempots = fixed_chempot.copy()
                        chempots.update({el.symbol:mu})
                    
            elif len(pd.elements) == 3: # 3-component PD case
                pdh = PDHandler(pd)
                res = pdh.get_phase_boundaries_chempots(target_comp, fixed_chempot)
                for el in list(res.values())[0]:
                    chempots_dict[el] = np.mean(np.array([mu[el] for mu in res.values()]))
                    
            elif len(pd.elements) > 3: 
                raise NotImplementedError('Not implemented for PD with more than 3 components')
            
            chempots = Chempots(chempots_dict)
            chempots_abs = chempots.get_absolute(pdh.mu_refs)
            
            if extrinsic_chempots_range:
                for el in extrinsic_chempots_range:
                    mu_el_O_poor,mu_el_O_rich = extrinsic_chempots_range[el][0], extrinsic_chempots_range[el][1]
                    chempots_abs[el] = mu_el_O_poor + ((mu_el_O_rich - mu_el_O_poor)/npoints)*i 
                        
            if get_pressures_as_strings:
                p = "%.1g" % p
                p = str(p)
            else:
                p = float("{:.3e}".format(p))
            reservoirs[p] = chempots_abs
    
            
        return PressureReservoirs(reservoirs,temperature,phase_diagram=pd,are_chempots_delta=False)
                                

    def get_oxygen_pressure_reservoirs(self,oxygen_ref,temperature=None,pressure_range=(-20,10),npoints=50,get_pressures_as_strings=False):
        """
        Get PressureReservoirs object for oxygen starting from the reference value.

        Parameters
        ----------
        oxygen_ref : (float)
            Absolute chempot of oxygen at 0K.
        temperature : (float), optional
            Temperature.
        pressure_range : (tuple), optional
            Exponential range in which to evaluate the partial pressure . The default is from 1e-20 to 1e10.
        npoints : (int), optional
            Number of data points to interpolate the partial pressure with. The default is 50.
        get_pressures_as_strings : (bool), optional
            Get pressure values (keys in the Reservoirs dict) as strings. The default is set to floats.

        Returns
        -------
        PressureReservoirs object

        """
        reservoirs = {}
        mu_refs = Chempots({'O':oxygen_ref})
        temperature = temperature if temperature else self.temperature
        partial_pressures = np.logspace(pressure_range[0],pressure_range[1],num=npoints,base=10)
        mu_standard = self.oxygen_standard_chempot(temperature)
        for p in partial_pressures:
            mu = {}
            muO = self.chempot_ideal_gas(mu_standard,temperature=temperature,partial_pressure=p)
            mu.update({'O':oxygen_ref + muO})
            if get_pressures_as_strings:
                p = "%.1g" % p
                p = str(p)
            else:
                p = float("{:.3e}".format(p))
            reservoirs[p] = Chempots(mu)
        
        return PressureReservoirs(reservoirs,temperature,phase_diagram=None,
                                  mu_refs=mu_refs,are_chempots_delta=False)
    
    
    def oxygen_standard_chempot(self,temperature=None):
        """
        Get value of oxygen delta mu standard (mu_0(T,Po)) at a speficic temperature
        The data is taken from the following work: 
            Reuter and Scheffler, “Composition, Structure, and Stability of RuO 2 ( 110 ) as a Function of Oxygen Pressure.”
        The value at a specific temperature is extracted from a linear fitting of the behaviour of mu_O with Temperature.
        
        Parameters
        ----------
        temperature : (float)

        Returns
        -------
        Chemical potential at standard p0 at given temperature.

        """
        temperature = temperature if temperature else self.temperature
        T = np.array([100,200,300,400,500,600,700,800,900,1000])
        mu_0 = np.array([-0.08, -0.17,-0.27,-0.38,-0.50,-0.61,-0.73,-0.85,-0.98,-1.10])
        
        coef = np.polyfit(T,mu_0,1)
        poly1d_fn = np.poly1d(coef)
        return poly1d_fn(temperature)            
        
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 28 14:23:05 2023

@author: villa
"""

import os
import warnings
import copy

from monty.json import MSONable

from pynter import SETTINGS
from pynter.tools.utils import grep_list


class Slurm(dict,MSONable):
    
    def __init__(self,load_defaults=True,**kwargs):
        """
        Object to handle sbatch commands in Slurm. Behaves like a dictionary.

        Parameters
        ----------
        **kwargs : 
            All possible sbatch arguments.
        """
        super().__init__()
        self._arguments, self._arguments_legend = read_possible_slurm_arguments()
        
        defaults = SETTINGS['job_settings']['slurm']
        
        settings = {}
        if load_defaults:
            for key,value in defaults.items():
                settings[key] = value
        
        for key,value in kwargs.items():
            if key in self.arguments_legend.keys():
                key = self.arguments_legend[key]
            settings[key] = value
        
        settings = dict(sorted(settings.items(), key=lambda item: item[0]))
        # for k,v in settings.items():
        #     self[k] = v # u
        self.update(settings)
        
    
    def __setitem__(self,key,value):
        if key not in self.arguments:
            warnings.warn(f'"{key}" is not listed as possible slurm argument')
        else:
            super().__setitem__(key,value)
        return
    
    def as_dict(self):
        d = dict(self)
        d["@module"] = type(self).__module__
        d["@class"] = type(self).__name__
        return d        
    
    def copy(self):
        return copy.deepcopy(self)
    
    def update(self, other=None, **kwargs):
        if other is not None:
            if hasattr(other, "items"):
                for key,value in other.items():
                    self.__setitem__(key, value)
            else:
                for key, value in other:
                    self.__setitem__(key, value)
        return
    
    @classmethod
    def from_dict(cls,d):
        return cls(**{k: v for k, v in d.items() if k not in ["@module", "@class"]})
    
    
    @property
    def arguments(self):
        """
        List of all possible sbatch arguments
        """
        return self._arguments
    
    @property
    def arguments_legend(self):
        """
        Dictionary which maps the short version (-<letter>) of the
        arguments with the long one (--<word>)
        """
        return self._arguments_legend
    
    
    @staticmethod
    def from_bash_script(input_string):
        """
        Initialize object from a bash script
        """
        lines = input_string.split('\n')
        kwargs = {}
        string = '#SBATCH'
        target_lines = grep_list(string,lines)
        for line in target_lines:  
            words = line.split(' ')
            if '-' in line and '--' not in line:
                if len(words) > 2:
                    pair = words[1].strip('-') , words[2]
                else:
                    pair = words[1].strip('-')
            elif '--' in line:
                pair =  words[1].strip('--').split('=')
            key = pair[0]
            value = pair[1] if len(pair) > 1 else ''
            if value.isdigit():
                value = int(value)
            kwargs.update({key:value})
            
        return Slurm(**kwargs)
     

    def get_bash_script_lines(self):
        """
        Get list of lines containing #SBATCH arguments for bash script
        """
        lines = []
        lines.append('#!/bin/sh\n')
        for key,value in self.items():
            if value is not None:
                printed_value = f'={value}' if value else '' 
                lines.append(f'#SBATCH --{key}{printed_value}\n')
        return lines
                
                
                

def read_possible_slurm_arguments():
    """
    Read possible slurm arguments from txt file and map (when present) their short versions 
    (-<letter>) with their long versions (--<word>).

    Returns
    -------
    arguments : (list)
        List containing all possible arguments.
    arguments_legend : (dict)
        Dictionary containing argument mapping.
    """
    path = os.path.abspath(__file__).strip(os.path.basename(__file__))
    filename = 'slurm_arguments.txt'
    with open(os.path.join(path,filename),'r') as file:
        lines = file.readlines()
  
    arguments = []
    arguments_legend = {}
    for line in lines:
        line = line.strip('\n')
        elements = line.split(',')
        arg = elements[0]
        arguments.append(arg)
        if len(elements) > 1:
            arg_short = elements[1]
            arguments_legend.update({arg_short:arg})

    return arguments, arguments_legend


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 11 16:33:19 2023

@author: villa
"""


import importlib

from pynter.testing.core import PynterTest
from pynter.testing.slurm import JobSettingsTest

class JobTest(PynterTest):
    """
    Contains methods to test Job objects
    """
    
    def __init__(self,jobclass='VaspJob'):
        """
        Specify which class the job object belongs to. Currently only "VaspJob" is available.
        """
        if jobclass == 'VaspJob':
            module = importlib.import_module("pynter.testing.vasp")
            self.cls = getattr(module,'VaspJobTest')
        elif jobclass == 'VaspNEBJob':
            module = importlib.import_module("pynter.testing.vasp")
            self.cls = getattr(module,'VaspNEBJobTest')
        super().__init__()
        
    def assert_inputs_equal(self,job1,job2,**kwargs):
        self.cls().assert_inputs_equal(job1,job2,**kwargs)
        return

    def assert_outputs_equal(self,job1,job2,**kwargs):
        self.cls().assert_outputs_equal(job1,job2,**kwargs)
        return
        
    def assert_settings_equal(self,job1,job2):
        JobSettingsTest().assert_job_settings_equal(job1.job_settings,job2.job_settings)
        
    def assert_job_equal(self,job1,job2,kwargs_input={},kwargs_output={}):
        self.assert_inputs_equal(job1,job2,**kwargs_input)
        if job1.outputs or job2.outputs:
            self.assert_outputs_equal(job1,job2,**kwargs_output)
        self.assert_settings_equal(job1,job2)
            
            
class DatasetTest(PynterTest):
    
    def assert_dataset_equal(self,ds1,ds2,kwargs_input={},kwargs_output={}):
        """
        Check if jobs in dataset are equal. Jobs are sorted before they are checked.
        """
        ds1.sort_jobs()
        ds2.sort_jobs()
        for i in range(len(ds1.jobs)):
            job1,job2 = ds1.jobs[i], ds2.jobs[i]
            JobTest(jobclass=job1.jobclass).assert_job_equal(job1, job2,kwargs_input,kwargs_output)
            
    def assert_jobs_equal(self,jobs1,jobs2,kwargs_input={},kwargs_output={}):
        """
        Check if lists of jobs are equal.
        """
        assert len(jobs1) == len(jobs2), "Job lists have different lengths"
        for i in range(0,len(jobs1)):
            j1, j2 = jobs1[i], jobs2[i]
            assert j1.jobclass == j2.jobclass, "Jobs belong to different classes" 
            JobTest(jobclass=j1.jobclass).assert_job_equal(j1, j2,kwargs_input,kwargs_output)
            
            
            
            
            
            




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  5 16:24:40 2023

@author: villa
"""

import os
import yaml

from pynter import get_config_from_default_file, get_vasp_defaults_from_default_file

def setup_config(subparsers):
    parser_configure = subparsers.add_parser('configure',help='Set up pynter configuration files')
    
    parser_configure.add_argument('-e','--exclude',help='Exclude file ("config.yml" or "vasp.yml")',
                               default=None,metavar='',dest='exclude')
    parser_configure.add_argument('-i','--info',action='store_true',help='Print configuration files paths',
                               default=False,dest='info')
    parser_configure.set_defaults(func=create_config)
    
    return
  

def create_config(args):
    run_config(exclude=args.exclude,info=args.info)
    return
  
def run_config(exclude=None,info=False):
    
    filename_config = 'config.yml'
    filename_vasp = 'vasp.yml'
    filepath = os.path.join(os.getenv("HOME"),'.pynter')
    
    # print configuration files path
    if info:
        print('Path of general configuration file: %s' %(os.path.join(filepath,filename_config)))
        print('Path of vasp configuration file: %s' %(os.path.join(filepath,filename_vasp)))
        return

    # Set config.yml
    if exclude != filename_config:
        print('\nSet up configuration file for pynter...\n')
        print('Path of configuration file is: %s' %filepath)
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        
        filename = filename_config
        
        print('Begin setup\n')
        print('HPC setup:')
        hostname = input('Hostname: ')
        localdir = input('Local calculation directory: ')
        localdir = os.path.abspath(localdir) if localdir else None
        workdir = input('Remote calculation directory: ')
        workdir = os.path.normpath(workdir) if workdir else None
        
        print('\nDefault settings for job script setup: ')
        project_id = input('Project ID: ')
        email = input('email address (Default None): ')
        path_exe = input('Path of default executable: ')
        job_script_filename = input('Job script filename (default job.sh): ')
        
        API_KEY = input('\nMaterials Project API_KEY (default None):')
        
        print('\nDatabase configuration:')
        host = input('Host (default: 127.0.0.1 ): ')
        port = input('Port (default: 27017): ')
        database = input('Database (default: pynter): ')
        collection = input('Collection (default vasp): ')
        
        
        job_script_filename = job_script_filename if job_script_filename else 'job.sh'
        host = host if host else '127.0.0.1'
        port = port if port else 27017
        database = database if database else 'pynter'
        collection = collection if collection else 'vasp'
        
        config = get_config_from_default_file()
        config.update({'HPC': 
                          {'hostname': hostname,
                          'localdir': localdir,
                          'workdir': workdir}})
        config.update({'API_KEY':API_KEY})
        
        config['job_settings']['filename'] = job_script_filename
        config['job_settings']['path_exe'] = path_exe
        config['job_settings']['slurm']['account'] = project_id
        config['job_settings']['slurm']['mail-user'] = email
        
        config.update({'dbconfig': 
                          {'vasp': 
                                {'collection': collection,
                                'database': database,
                                'host': host,
                                'password': None,
                                'port': port,
                                'user': None}}})
                
        with open(os.path.join(filepath,filename),'w+') as f:
            yaml.dump(config,f) 
            
        print(f'Configuration file {filename} written in {filepath}\n')    
        
        
    # set vasp.yml
    if exclude != filename_vasp:
    
        filename = filename_vasp
        vasp_default = get_vasp_defaults_from_default_file()
        
        with open(os.path.join(filepath,filename),'w+') as f:
            yaml.dump(vasp_default,f) 
        
        print(f'Vasp default file {filename} written in {filepath}')  
















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  3 19:25:00 2020

@author: villa
"""
import os
import os.path as op
from glob import glob
from shutil import copyfile
from pynter.automations.core import CommandHandler, Automation
from pymatgen.io.vasp.inputs import Kpoints, Incar
from pymatgen.io.vasp.outputs import Vasprun, Oszicar, Outcar
from pynter.tools.utils import grep


class Base(Automation):
    """
    Subclass of Automation that contains methods for automations of VASP calculations. Works as base class for schemes.
    """ 

    def __init__(self,job_script_filename=None,status_filename='exit_status.txt',path=None):
        super().__init__(job_script_filename,status_filename,path)


    @property
    def vasprun(self):
        return self.read_vasprun()
    
    
    def compare_kpoints(self,dir1=None,dir2=None):
        """
        Compare KPOINTS files in 2 different folders

        Parameters
        ----------
        dir1 : (str), optional
            First path. The default is None. If None "path" attribute is used.
        dir2 : (str), optional
            Second path. The default is None. If None the next step path is used (determined with "get_next_step()").

        Returns
        -------
        kpoints_are_same : (bool)
            True if KPOINTS files are the same, False if they're different.
        """        
        dir1 = dir1 if dir1 else self.path
        dir2 = dir2 if dir2 else self.get_next_step()       
        current_kpoints = Kpoints().from_file(os.path.join(dir1,'KPOINTS')).as_dict()        
        if self.get_next_step():
            next_kpoints = Kpoints().from_file(os.path.join(dir2,'KPOINTS')).as_dict()
            kpoints_are_same = True if current_kpoints == next_kpoints else False
        else:
            kpoints_are_same = None            
        return kpoints_are_same
    

    def convergence(self,path=None):
        """
        Check electronic and ionic convergence by reading "vasprun.xml" file with Pymatgen.\n
        If reading of vasprun failes, returns False for electronic and ionic convergence.

        Parameters
        ----------
        path : (str), optional
            Path of "vasprun.xml" file. The default is None. If None "path" attribute is used.        

        Returns
        -------
        converged_electronic : (bool)
            Electronic convergence reached.
        converged_ionic : (bool)
            Ionic convergence reached.
        """                
        path = path if path else self.path
        converged_electronic = False
        converged_ionic = False
        try:
            vasprun = self.read_vasprun(path=path)
            converged_electronic = vasprun.converged_electronic
            converged_ionic = vasprun.converged_ionic
        except:
            print('"vasprun.xml" could not be read, calculation probably did not converge')
            pass
            
        return converged_electronic, converged_ionic
       

    def find_NEB_dirs(self,path=None):
        """
        Find directories of images for NEB calculations. Directories are selected if all characters in the
        directory name are digits.
        """
        dirs = []
        path = path if path else self.path
        path = op.abspath(path)
        for d in os.walk(path):
            directory = d[0]
            image_name = op.relpath(directory,start=path)
            if all(c.isdigit() for c in list(image_name)): #check if folder is image (all characters in folder rel path need to be numbers)
                dirs.append(directory)
        dirs.sort()
        return dirs


    def limit_electronic_steps_reached(self):
        """
        Check if limit of electronic steps has been reached. This methods compares the "NELM" flag 
        in the INCAR file with the number of electronic steps performed in the calculation

        Returns
        -------
        bool
            True if limit reached.
        """
        steps = self.vasprun.ionic_steps
        n_electronic_steps = len(steps[-1]['electronic_steps'])
        nelm = self.vasprun.parameters['NELM']        
        return True if n_electronic_steps == nelm else False
        

    def limit_ionic_steps_reached(self):
        """
        Check if limit of ionic steps has been reached. This methods compares the "NSW" flag 
        in the INCAR file with the number of ionic steps performed in the calculation

        Returns
        -------
        bool
            True if limit reached.
        """
        n_steps = len(self.vasprun.ionic_steps)
        nsw = self.vasprun.parameters['NSW']                
        return True if n_steps == nsw else False

 
    def read_oszicar(self,path=None):
        """
        Get Pymatgen OSZICAR object by reading "OSZICAR" file        

        Parameters
        ----------
        path : (str), optional
            Path of "OSZICAR" file. The default is None. If None "path" attribute is used.

        Returns
        -------
        Oszicar object
        """
        path = path if path else self.path
        return Oszicar(os.path.join(path,'OSZICAR'))


    def read_outcar(self,path=None):
        """
        Get Pymatgen OSZICAR object by reading "OUTCAR" file        

        Parameters
        ----------
        path : (str), optional
            Path of "OUTCAR" file. The default is None. If None "path" attribute is used.

        Returns
        -------
        Outcar object
        """
        path = path if path else self.path
        return Outcar(os.path.join(path,'OUTCAR'))

    
    def read_vasprun(self,path=None):
        """
        Get Pymatgen Vasprun object by reading "vasprun.xml"        

        Parameters
        ----------
        path : (str), optional
            Path of "vasprun.xml" file. The default is None. If None "path" attribute is used.

        Returns
        -------
        Vasprun object
        """
        path = path if path else self.path
        return Vasprun(os.path.join(path,'vasprun.xml'))





class Schemes(Base):
    
    def __init__(self,job_script_filename=None,status_filename='exit_status.txt',path=None,status=[],**kwargs):
        """
        Class to organize automation schemes for Vasp

        Parameters
        ----------
        vasp_automation : (VaspAutomation object)
            VaspAutomation object.
        status : (list), optional
            List of lines containing the report of the automation. The default is [].
        **kwargs :
            If None the default arguments created by the CommandHandler class are used.
        """
        super().__init__(job_script_filename,status_filename,path)
        
        self.status = status

        if kwargs: # custom args
            for key, value in kwargs.items():
                setattr(self,key,value)
        
        else:
            args = CommandHandler().vasp_args() # default args
            for key, value in args.__dict__.items():
                setattr(self,key,value)

    
    def compare_next_step_kpoints(self):
        """
        Compare KPOINTS file in current and next step directory 

        Returns
        -------
        kpoints_are_same : bool
            True if KPOINTS files are the same.
        """
        kpoints_are_same = self.compare_kpoints()
        if kpoints_are_same:
            self.status.append('KPOINTS of current and next step are the same')
        else:
            self.status.append('KPOINTS of current and next step are different')
        return kpoints_are_same 
            
      
    def check_convergence(self):
        """
        Get information on electronic and ionic convergence and adds it to status.

        Returns
        -------
        conv_el : (bool)
            Electronic convergence reached.
        conv_ionic : (bool)
            Ionic convergence reached.

        """        
        conv_el, conv_ionic = self.convergence()
        self.status.append('Job exit. Analysing "vasprun.xml"... ')
        self.status.append(f'\nElectronic convergence: {conv_el}')
        self.status.append(f'Ionic convergence: {conv_ionic}\n') 
        if conv_el == False and conv_ionic == False:
            self.status.append('Calculation probably did not converge, control manually')
        return conv_el,conv_ionic


    def next_step_relaxation_schemes(self):
        """
        If next step exists, transfer output files from the current calculation to 
        the next step calculation and submits the new job. \n
        If "check_kpoints" attribute is set to True, CHGCAR and WAVECAR are copied 
        only if KPOINTS file of current and next step are the same. \n
        CHGCAR,WAVECAR and CONTCAR can be copied if their related attributes ("chgcar","wavecar","contcar") are set to True. \n
        This can be done either manually or by including the commands (-C or --chgcar, -W or --wavecar, -c or --contcar) in the command line.
        """
        if self.get_next_step():
            # if True copy CHGCAR and WAVECAR only if KPOINTS are the same, else copy them in any case
            if self.check_kpoints:
                if self.compare_next_step_kpoints():
                    if self.chgcar:
                        self.transfer_chgcar()
                    if self.wavecar:
                        self.transfer_wavecar()                       
            else:
                if self.chgcar:
                    self.transfer_chgcar()
                if self.wavecar:
                    self.transfer_wavecar()                
            # if True copy CONTCAR to next step POSCAR
            if self.contcar:
                self.transfer_contcar_to_poscar()
                                                  
            self.submit_job()
            self.status.append('\nNext step calculation submitted in dir "../%s"' %os.path.basename(self.get_next_step()))            
        else:
            self.status.append('\nNo next step found, no other calculation submitted')        


    def print_status(self):
        """
        Print status lines on terminal
        """
        print('\n'.join(self.status))


    def step_limits_reached(self):
        """
        Check if electronic or ionic steps limit has been reached. \n
        Electronic steps limit is taken from the "NELM" flag in the INCAR file. \n
        Ionic steps limit is taken from the "NSW" flag in the INCAR file

        Returns
        -------
        el_limit_reached : (bool)
        
        ionic_limit_reached : (bool)
        """
        if self.limit_electronic_steps_reached():
            el_limit_reached = True
            self.status.append('Maximum limit of electronic steps has been reached')
        else:
            el_limit_reached = False
        if self.limit_ionic_steps_reached():
            ionic_limit_reached = True
            self.status.append('Maximum limit of ionic steps has been reached')
        else:
            ionic_limit_reached = False
        return el_limit_reached, ionic_limit_reached
                    

    def resubmit_if_step_limits_reached(self):
        """
        To use if calculation is not converged. Does not perform actions in case the number of files named "out.*"
        in the working directory is >= 2. The idea is that if the calculation hasn't converged after reaching the limit
        twice, is better to control manually. \n
        
        Check if electronic or ionic steps limit has been reached. \n
        In case electronic steps limit is reached, resubmits calculation without doing anything. \n
        In case ionic steps limit is reached copies CONTCAR to POSCAR and resubmits calculation.
        """
        
        number_of_out_files = len(glob(os.path.join(self.path,'out*')) )
        if number_of_out_files < 2:
            self.status.append('Checking if limit of electronic or ionic steps have been reached...')
            el_limit, ionic_limit = self.step_limits_reached()
            
            if el_limit==True and ionic_limit==False:
                self.submit_job(job_script_path = self.path)
                self.status.append('Resubmitting calculation. Be sure "ISTART" in your INCAR is set to 1 or default.')
            
            if el_limit==False and ionic_limit==True:
                copyfile(os.path.join(self.path,'CONTCAR'),os.path.join(self.path,'POSCAR'))
                self.submit_job(job_script_path = self.path)
                self.status.append('Resubmitting calculation. Be sure "ISTART" in your INCAR is set to 1 or default.')
            
            if el_limit==False and ionic_limit==False:
                self.status.append('Problem is not due to steps limit being reached, controll manually...')
        else:
            self.status.append('Two out.* files are already present, control manually...')


    def transfer_chgcar(self):
        """
        Copy CHGCAR file to next step directory
        """
        self.copy_files_to_next_step_dir('CHGCAR')
        self.status.append('CHGCAR copied in dir "../%s"' %os.path.basename(self.get_next_step()))

           
    def transfer_contcar_to_poscar(self):
        """
        Copy CONTCAR file to POSCAR file in next step directory
        """
        self.copy_files_to_next_step_dir(('CONTCAR','POSCAR'))
        self.status.append('CONTCAR copied in POSCAR in dir "../%s"' %os.path.basename(self.get_next_step()))
        return

    
    def transfer_wavecar(self):
        """
        Copy WAVECAR file to next step directory
        """
        self.copy_files_to_next_step_dir('WAVECAR')
        self.status.append('WAVECAR copied in dir "../%s"' %os.path.basename(self.get_next_step()))        
        return
    
                
    def write_status(self,filename=None,path=None):
        """
        Write status lines to file

        Parameters
        ----------
        filename : (str), optional
            Filename. The default is None. If None "status_filename" attribute is used.
        path : (str), optional
            Path of file to write. The default is None. If None "path" attribute is used.
        """

        filename = filename if filename else self.status_filename
        path = path if path else self.path
        if filename:
            with open(filename,'w') as f:
                f.write('\n'.join(self.status))
                print('Automation report written in "%s" file in directory "%s"' %(filename,path))
        else:
            print('Writing status file is disabled, if needed please set "status_filename" kwarg or "--status" optional argument on command line')
            
            
class NEBSchemes(Schemes):


    def clean_NEB_dirs(self,printout=False):
        """
        Clean image directories for NEB calculation. All files but CHGCAR,WAVECAR,POSCAR,CONTCAR and OUTCAR are removed.
        """
        image_dirs = self.find_NEB_dirs()
        for d in image_dirs:
            files = [w[2] for w in os.walk(d)][0]
            for f in files:
                if f not in ('CHGCAR','WAVECAR','POSCAR','OUTCAR','CONTCAR'):
                    os.remove(op.join(d,f))
                    message = f'Removed {op.join(d,f)} \n'
                    if printout:
                        print(message)
                    self.status.append(message)                            
        return
    

    def check_ionic_relaxation_from_outfile(self):
        """
        Useful for NEB because Pymatgen fails to read vasprun file for NEB calculations.
        This function reads the outfile with highest number in the dir and checks for the 
        string: "reached required accuracy - stopping structural energy minimisation". 
        """
        outfiles = glob(os.path.join(self.path,'out*'))
        if outfiles:
            outfiles.sort()
            outfile = outfiles[-1]
            lines = grep('reached required accuracy - stopping structural energy minimisation',outfile)
            if lines:
                print('"reached required accuracy - stopping structural energy minimisation" found in %s' %outfile)
                self.status.append('"reached required accuracy - stopping structural energy minimisation" found in %s' %outfile)
                return True
        else:
            return False
        

    def check_limit_ionic_steps_oszicar(self,printout=False):
        """
        Function to check if limit of ionic steps is reached from OSZICAR file. Reading of vasprun.xml
        failes for NEB calculations. This function checks if the limit is reached from OSZICAR file
        in all images directories. The reference is the NSW tag in the INCAR file in the parent dir.
        """
        limit_reached = True
        image_dirs = self.find_NEB_dirs()
        for d in image_dirs:
            if d != image_dirs[0] and d != image_dirs[-1]:
                image_name = op.basename(d)
                n_steps = len(self.read_oszicar(path=d).ionic_steps)
                nsw = Incar.from_file(op.join(self.path,'INCAR'))['NSW'] # check NSW from INCAR in parent directory
                if printout:
                    self.status.append(f'{image_name}: number of ionic steps:{n_steps}, INCAR["NSW"]:{nsw}')
                if nsw != n_steps:
                    limit_reached = False
        if limit_reached:
            self.status.append('Limit of ionic steps has been reached')
        return limit_reached
            
            
    def check_preconvergence_images(self):
        """
        Check if all SCF calculations of the images are converged
        """
        if self.is_preconvergence():
            convergence = True
            image_dirs = self.find_NEB_dirs()
            for d in image_dirs:
                conv_el, conv_ionic = self.convergence(path=d)
                if conv_el == False or conv_ionic == False:
                    convergence = False
                    self.status.append(f'convergence in {d}: False')
                else:
                    self.status.append(f'convergence in {d}: True')
        else:
            raise ValueError("Current main folder doesn't contain preconvergence calculations of the images")
        
        return convergence
                

    def copy_images_next_step(self):
        """
        Copy CHGCAR,WAVECAR and CONTCAR in POSCAR of respective images in next step folders
        """
        next_step_path = self.get_next_step()
        image_dirs = self.find_NEB_dirs()
        for d in image_dirs:
            image_name = op.basename(d)
            files = [w[2] for w in os.walk(d)][0]
            for f in files:
                if f in ('CHGCAR','WAVECAR','OUTCAR'):
                    source = op.join(d,f)
                    dest = op.join(next_step_path,image_name,f)
                    copyfile(source,dest)
                    self.status.append(f'{f} copied in {dest}')
                if f == 'CONTCAR':
                    source = op.join(d,'CONTCAR')
                    dest = op.join(next_step_path,image_name,'POSCAR')    
                    copyfile(source,dest)
                    self.status.append(f'{f} copied in {dest}')
        return
        

    def copy_images_next_step_and_submit(self):
        """
        Copy CHGCAR,WAVECAR and CONTCAR in POSCAR of respective images in next step folders 
        and submit calculation in next step.
        """
        self.copy_images_next_step()
        self.submit_job()
        self.status.append('Calculation in dir "%s" submitted' %os.path.basename(self.get_next_step()))
        return


    def is_NEB_job_finished(self):
        """
        Returns True if job is converged or if the limit of ionic steps has been reached
        """
        is_job_finished = False
        
        if self.check_ionic_relaxation_from_outfile():
            is_job_finished = True
        elif self.check_limit_ionic_steps_oszicar():
            is_job_finished = True
        
        if is_job_finished:
            self.status.append('NEB job is finished: converged or limit of ionic steps has been reached')
        return is_job_finished
                
                
    def is_preconvergence(self):
        """
        Check if current main folder contains images folders with preconvergence calculations.
        criteria for selection of preconvergence calculation is the presence of the INCAr file
        in the image folder.
        """
        is_preconvergence = False
        image_dirs = self.find_NEB_dirs()
        for d in image_dirs:
            files = [w[2] for w in os.walk(d)][0]
            if 'INCAR' in files:
                is_preconvergence = True
            elif is_preconvergence:
                print('Warning: Inconsistencies in images directories: INCAR file is present but not in all directories')
        return is_preconvergence
        
        
    
        
        
        
        


#!/usr/bin/env python

from pymatgen.ext.matproj import MPRester

from pynter import  get_cfgfile
from pynter import SETTINGS

try:
    API_KEY = SETTINGS['API_KEY']
except:
    cfgfile = get_cfgfile()
    raise KeyError('"API_KEY" needs to be present in the %s file' %cfgfile)


class MPDatabase:
    
    def __init__(self,mp_id=None):
        """
        Class to retrieve data from Materials Project database

        Parameters
        ----------
        mp_id : (str), optional
            Materials-ID. The default is None.
        """
        
        self.mp_id = mp_id if mp_id else None
        self.api_key = API_KEY
    
    @property
    def mp_rester(self):
        return MPRester(API_KEY)
        
        
    def get_data(self,data_type='vasp'):    
        """
        Flexible method to get any data using the Materials Project REST
        interface. Generally used by other methods for more specific queries.

        Format of REST return is *always* a list of dict (regardless of the
        number of pieces of data returned. The general format is as follows:

        [{"material_id": material_id, "property_name" : value}, ...]

        This is generally a call to
        https://www.materialsproject.org/rest/v2/materials/vasp/<prop>.
        See https://github.com/materialsproject/mapidoc for details.

        Parameters
        ----------
            data_type (str): Type of data to return. Currently can either be
                "vasp" or "exp".
        Returns:
            List of dictionaries with available MP data
        """
        with MPRester(API_KEY) as mpr:
            data = mpr.get_data(self.mp_id,data_type=data_type) 
        
        return data


    def get_entries(self,chemsys_formula_id_criteria,compatible_only=True,inc_structure='initial',
                    property_data=None,conventional_unit_cell=False,sort_by_e_above_hull=True):
        """
        Get a list of ComputedEntries or ComputedStructureEntries corresponding
        to a chemical system, formula, or materials_id or full criteria.

        Parameters
        ----------
            chemsys_formula_id_criteria (str/dict): A chemical system
                (e.g., Li-Fe-O), or formula (e.g., Fe2O3) or materials_id
                (e.g., mp-1234) or full Mongo-style dict criteria.
            compatible_only (bool): Whether to return only "compatible"
                entries. Compatible entries are entries that have been
                processed using the MaterialsProject2020Compatibility class,
                which performs adjustments to allow mixing of GGA and GGA+U
                calculations for more accurate phase diagrams and reaction
                energies.
            inc_structure (str): If None, entries returned are
                ComputedEntries. If inc_structure="initial",
                ComputedStructureEntries with initial structures are returned.
                Otherwise, ComputedStructureEntries with final structures
                are returned.
            property_data (list): Specify additional properties to include in
                entry.data. If None, no data. Should be a subset of
                supported_properties.
            conventional_unit_cell (bool): Whether to get the standard
                conventional unit cell
            sort_by_e_above_hull (bool): Whether to sort the list of entries by
                e_above_hull (will query e_above_hull as a property_data if True).

        Returns:
            List of ComputedEntry or ComputedStructureEntry objects.
        """
        with MPRester(API_KEY) as mpr:
            entries = mpr.get_entries(chemsys_formula_id_criteria,compatible_only,inc_structure,
                                   property_data,conventional_unit_cell,sort_by_e_above_hull)
        return entries
    
    
    def get_entries_from_compositions(self,compositions,lowest_e_above_hull=False,compatible_only=True,
                                      inc_structure='initial',property_data=None,
                                      conventional_unit_cell=False):
        """
        Get a dictionary with compositions (strings) as keys and a list of ComputedEntries 
        or ComputedStructureEntries as values.

        Parameters
        ----------
            compositions (list): List of strings with compositions.
            stable_only (bool): Get phase with lowest E above hull.
            compatible_only (bool): 
                Whether to return only "compatible"
                entries. Compatible entries are entries that have been
                processed using the MaterialsProject2020Compatibility class,
                which performs adjustments to allow mixing of GGA and GGA+U
                calculations for more accurate phase diagrams and reaction
                energies.
            inc_structure (str):
                If None, entries returned are
                ComputedEntries. If inc_structure="initial",
                ComputedStructureEntries with initial structures are returned.
                Otherwise, ComputedStructureEntries with final structures
                are returned.
            property_data (list):
                Specify additional properties to include in
                entry.data. If None, no data. Should be a subset of
                supported_properties.
            conventional_unit_cell (bool):
                Whether to get the standard conventional unit cell
            sort_by_e_above_hull (bool): Whether to sort the list of entries by
                e_above_hull (will query e_above_hull as a property_data if True).

        Returns:
            List of ComputedEntry or ComputedStructureEntry objects.
        """
        entries_dict = {}
        for comp in compositions:
            entries = self.get_entries(comp,compatible_only,inc_structure,property_data,
                                                  conventional_unit_cell,sort_by_e_above_hull=True)
            if lowest_e_above_hull:
                entries_dict[comp] = entries[0] # sorted by e_above_hull
            else:
                entries_dict[comp] = entries
        return entries_dict
    

    def get_structure(self,final=False,conventional_unit_cell=False):
        """
        Get a Structure corresponding to a material_id.

        Parameters
        ----------
            material_id (str): Materials Project material_id (a string,
                e.g., mp-1234).
            final (bool): Whether to get the final structure, or the initial
                (pre-relaxation) structure. Defaults to True.
            conventional_unit_cell (bool): Whether to get the standard
                conventional unit cell

        Returns:
            Structure object.
        """
        with MPRester(API_KEY) as mpr:
            structure = mpr.get_structure_by_material_id(self.mp_id,final=final,conventional_unit_cell=conventional_unit_cell).get_sorted_structure()
        return structure

    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 11 16:33:19 2023

@author: villa
"""

import unittest
from numpy.testing import assert_array_almost_equal


class StructureTest(unittest.TestCase):
    """
    Provides methods to test pymatgen Structure objects
    """
    def assert_Site_equal(self,site1,site2,decimal=3):
        assert_array_almost_equal(site1.frac_coords,site2.frac_coords,decimal=decimal)
        self.assertEqual(site1.specie.symbol,site2.specie.symbol)
        
    def assert_Lattice_equal(self,lattice1,lattice2):
        assert_array_almost_equal(lattice1,lattice2)
        
    def assert_Structure_equal(self,structure1,structure2):
        if len(structure1) != len(structure2):
            raise AssertionError
        else:
            for i in range(0,len(structure1)):
                self.assert_Site_equal(structure1[i], structure2[i])
        
            




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 23 11:19:30 2021

@author: villa
"""
from pymatgen.electronic_structure.plotter import BSPlotter,BSDOSPlotter,DosPlotter, BSPlotterProjected  #original pymatgen
#from pynter.vasp.pmg.pmg_electronic_structure_plotter import BSPlotter,BSDOSPlotter,DosPlotter, BSPlotterProjected #modified version 
from pymatgen.electronic_structure.bandstructure import BandStructure


def plot_bs(bs,**kwargs):
    """
    Get BS plot with pymatgen.

    Parameters
    ----------
    bs : 
        BandStructureSymmLine object, most likely generaten from Vasprun or BSVasprun.
    **kwargs : (dict)
        Arguments for the get_plot function in BSPlotter in pymatgen.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    plotter = BSPlotter(bs)
    plt = plotter.get_plot(**kwargs)
    
    return plt


def plot_bs_elt_projected(bs,legend=True,**kwargs):
    """
    Get element projected BS color plot with pymatgen. Note that vasprun.xml needs to
    be parsed with parse_projected_eigen set to True.

    Parameters
    ----------
    bs : 
        BandStructureSymmLine object, most likely generated from Vasprun or BSVasprun.
    legend: (bool or dict)
        Get colors legend. If is a dict it is passed as kwargs in plt.legend(**kwargs)
    **kwargs : (dict)
        Arguments for the get_elt_projected_plots_color function in BSPlotter in pymatgen.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    plotter = BSPlotterProjected(bs)
    plt = plotter.get_elt_projected_plots_color(**kwargs)
    
    import matplotlib.patches as mpatches
    colors = ['red','green','blue']
    patches = []
    if legend:
        if 'elt_ordered' not in kwargs.keys():
            elt_ordered = plotter._bs.structure.composition.elements
        for elt in elt_ordered:
            el = elt.symbol
            patches.append(mpatches.Patch(color=colors[elt_ordered.index(elt)],label=el))
    if type(legend) == dict:
        plt.legend(handles=patches,**legend)
    else:
        plt.legend(handles=patches)
    return plt


def plot_dos_bs(dos,bs,draw_fermi=True,**kwargs):
    """
    Plot DOS and BS with pymatgen's BSDOSPlotter class.

    Parameters
    ----------
    dos : 
        Dos object.
    bs : 
        BandStructureSymmLine object, most likely generated from Vasprun or BSVasprun..
    **kwargs : (dict)
        Arguments to pass to BSDOSPlotter class.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    if 'bs_projection' not in kwargs.keys(): # setting default projections to None
        kwargs['bs_projection'] = None
    if 'dos_projection' not in kwargs.keys():
        kwargs['dos_projection'] = None
    if draw_fermi:
        plt = BSDOSPlotter(**kwargs).get_plot(bs,dos) # ensure that also original pymatgen function works
    else:
        plt = BSDOSPlotter(**kwargs).get_plot(bs,dos,draw_fermi=draw_fermi)
        
        
    return plt


def plot_dos_dict(complete_dos,d,xlim=(-3,3),**kwargs):
    """
    Plot DOS dict with pymatgen. The dict is most likely generated with the methods of the
    pymatgen CompleteDos class.

    Parameters
    ----------
    complete_dos : 
        CompleteDos object.
    d : TYPE
        Dict of DOS.
    xlim : (tuple), optional
        Limits for x axis. The default is (-3,3).
    **kwargs : (dict)
        Arguments for DosPlotter in pymatgen.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    
    dos_plotter = DosPlotter(**kwargs)
    for k in d:
        dos_plotter.add_dos(k,d[k])
    eg = complete_dos.get_gap()
    plt = dos_plotter.get_plot(xlim=(xlim[0],eg+xlim[1]))
    
    return plt


def plot_element_dos(complete_dos,xlim=(-3,3),**kwargs):
    """
    Plot element resolved DOS with pymatgen

    Parameters
    ----------
    complete_dos : 
        CompleteDos object.
    xlim : (tuple), optional
        Limits for x axis. The default is (-3,3).
    **kwargs : (dict)
        Arguments for DosPlotter in pymatgen.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    d = complete_dos.get_element_dos()        
    plt = plot_dos_dict(complete_dos, d, xlim, **kwargs)
    
    return plt


def plot_spd_dos(complete_dos,xlim=(-3,3),**kwargs):
    """
    Plot spd resolved DOS with pymatgen

    Parameters
    ----------
    complete_dos : 
        CompleteDos object.
    xlim : (tuple), optional
        Limits for x axis. The default is (-3,3).
    **kwargs : (dict)
        Arguments for DosPlotter in pymatgen.

    Returns
    -------
    plt : 
        Matplotlib object.
    """
    d = complete_dos.get_spd_dos()        
    plt = plot_dos_dict(complete_dos, d, xlim, **kwargs)
    
    return plt


def plot_dos_bs_custom(bs,dos,ylim=(-4,7)):
    
    from pynter.vasp.pmg_electronic_structure_plotter import BSPlotter, DosPlotter
    # PLOT BS
    plt = BSPlotter(bs).get_plot(ylim=ylim,get_subplot=True)
    plt.xticks(fontsize=25)
    plt.yticks(fontsize=25)
        
    # PLOT DOS
    partial_dos = dos.get_spd_dos()
    dos_plotter = DosPlotter()
    dos_plotter.add_dos('Total_dos',dos)
    for orbital in partial_dos:
        dos_plotter.add_dos(orbital,partial_dos[orbital])
    plt = dos_plotter.get_plot(xlim=ylim,get_subplot=True)    

    # modify pymatgen output to flip graph
    ax = plt.gca()
    ymax = 0
    fermi_lines = []
    for i in range(0,len(ax.lines)):
        x = ax.lines[i].get_ydata()
        y = ax.lines[i].get_xdata()
        for j in range(0,len(x)):
            if x[j] > ylim[0] and x[j] < ylim[1]:
                if x[j] > ymax:
                    ymax = x[j]
        ax.lines[i].set_xdata(x)
        ax.lines[i].set_ydata(y)
        if ax.lines[i].get_linestyle() == '--':
            fermi_lines.append(ax.lines[i])
            
    ax.lines = [l for l in ax.lines if l not in fermi_lines] 
    ylim = ax.get_xlim()
    ax.set_xlim()
    ax.set_ylim()
    ax.set_xlim(0,ymax)
    ax.set_ylim(ylim)
    ax.set_xlabel('Density of states',size=25)
    ax.set_yticks([])
    ax.set_ylabel(None)
    plt.xticks(fontsize=25)

    fig = plt.gcf()
    fig.tight_layout()

    return plt




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  2 11:27:24 2023

@author: villa
"""

from pymatgen.io.vasp.inputs import Poscar


def setup_mp_database(subparsers):
    
    parser_sub = subparsers.add_parser('MP',help='Interface with Materials Project database')
    
    parser_sub.add_argument('-id','--mp-id',help='Materials project ID',required=True,type=str,metavar='',dest='mp_id')
    parser_sub.add_argument('-s','--structure',help='print Structure',action='store_true',required=False,default=False,dest='print_structure')
    parser_sub.add_argument('-P','--poscar',help='create POSCAR',action='store_true',required=False,default=False,dest='write_poscar')
    parser_sub.add_argument('-c','--conv-cell',help='Get conventional unit cell (default: %(default)s)',action='store_true',
                        required=False,default=False,dest='get_conventional')
    parser_sub.add_argument('-f','--final-structure',help='Get final structure (default: %(default)s)',action='store_true',
                        required=False,default=False,dest='get_final')     
    parser_sub.add_argument('-n','--filename',help='Filename for POSCAR (default: %(default)s)',required=False,
                        default='POSCAR',type=str,metavar='',dest='filename')
    
    parser_sub.set_defaults(func=run_mp_database)
    return
    

def run_mp_database(args):
    from pynter.tools.materials_project import MPDatabase
    mp = MPDatabase(mp_id=args.mp_id)
    structure = mp.get_structure(final=args.get_final,conventional_unit_cell=args.get_conventional)
    
    if args.print_structure:
        print(structure)
    if args.write_poscar:
        Poscar(structure).write_file(args.filename)
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 11:12:09 2020

@author: villa
"""

import warnings
from matgendb.creator import VaspToDbTaskDrone
from pynter import SETTINGS

dbconfig = SETTINGS['dbconfig']

# for VaspJob we use pymatgen-db interface
class VaspJobDrone(VaspToDbTaskDrone):
    """
    Subclass of VaspToDbTaskDrone in pymatgen-db (matgendb) package.
    To set up the database for VaspJob we use the pymatgen-db package which already provides a useful
    interface to the other pymatgen objects that are already used (like Structure or ComputedEntry).
    This class helps to interface a VaspJob object with the database.
    For documentation on input args (https://github.com/materialsproject/pymatgen-db/blob/master/matgendb/creator.py).
    Warning: 'use_full_uri' has been set to False here to allow interface with VaspJob objects.
    """
    def __init__(self, job, **kwargs):
        
        for k,v in dbconfig['vasp'].items():
            if k not in kwargs:
                kwargs[k] = v
                   
        super().__init__(use_full_uri=False,**kwargs)

        self.job = job
        self._path = self.job.path
    
    
    def assimilate_job(self,check_convergence=True):
        """
        Insert VaspJob intro database.
        """
        if check_convergence:
            if self.job.is_converged:
                doc = self.get_task_doc_from_files()     
                self._insert_doc(doc)
            else:
                warnings.warn(f'VaspJob "{self.job.name}" is not converged and will not be added into the database',UserWarning)
        else:
            doc = self.get_task_doc_from_files()     
            self._insert_doc(doc)
        return
        
    
    def get_task_doc_from_files(self):
        """
        Get entire task doc from "vasprun.xml" in job path.
        """
        doc = self.get_task_doc(self._path)
        doc['job_settings'] = self.job.job_settings
        doc['job_script_filename'] = self.job.job_script_filename
        doc['job_name'] = self.job.name
        doc['is_converged'] = self.job.is_converged
        return doc
    
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 31 14:50:11 2023

@author: villa
"""
import os

from pymatgen.core.composition import Composition
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.analysis.phase_diagram import PhaseDiagram

from pynter import SETTINGS
from pynter.phase_diagram.chempots import PDHandler, Reservoirs
from pynter.tools.utils import save_object_as_json


def setup_phase_diagram(subparsers):
    
    parser_sub = subparsers.add_parser('phase-diagram',help='Generate phase diagrams with pymatgen from the Materials Project database')
    
    parser_sub.add_argument('-e','--elements',action='append',help='Components of the phase diagram, provide as "el1-el2" or -e el1 -e el2',
                            type=str,metavar='',dest='elements')
    parser_sub.add_argument('-c','--chempots',help='Print boundary chempots of target phase',required=False,default=False,type=str,metavar='',
                            dest='composition_chempots')
    parser_sub.add_argument('-sc','--save-chempots',action='store_true',help='Save chempots in Reservoirs object as json',required=False,default=False,dest='savechempots')
    parser_sub.add_argument('-p','--plot-PD',action='store_true',help='Plot phase diagram',required=False,default=False,dest='plot_pd')
    parser_sub.add_argument('-ps','--plot-stability',action='append',help='Plot stability diagram, provide elements to be on axis as "el1-el2" or -ps el1 -ps el2'
                            ,required=False,type=str,dest='elements_stability')
    parser_sub.add_argument('-j','--json',action='store_true',help='Save PhaseDiagram as json',required=False,default=False,dest='savejson')
    parser_sub.add_argument('-s','--savefig',action='store_true',help='Save matplotlib figure as pdf',required=False,default=False,dest='savefig')
    parser_sub.set_defaults(func=run_phase_diagram)
    return


def run_phase_diagram(args):
    elements = args.elements[0].split('-') if len(args.elements)==1 else args.elements
    plt = None
    pd = get_phase_diagram(elements)
    print('\n',pd)
    
    if args.composition_chempots:
        composition = Composition(args.composition_chempots)
        chempots = get_chempots(pd,composition)
        reservoirs = Reservoirs(chempots,pd,are_chempots_delta=False)
        print('\n',f'Chemical potentials of {args.composition_chempots}:\n',reservoirs)
        if args.savechempots:
            fname = 'reservoirs_boundary_' + args.composition_chempots + '.json'
            path = os.path.join(os.getcwd(),fname)
            reservoirs.to_json(path)
    
    if args.plot_pd:
        plt = plot_pd(pd)
        if args.savefig:
            fname = 'PD_' + '-'.join(elements) + '.pdf'
            path = os.path.join(os.getcwd(),fname)
            plt.savefig(path,bbox_inches='tight')
        else:
            plt.show()
    if args.elements_stability:
        elements_stability = args.elements_stability[0].split('-') if len(args.elements_stability)==1 else args.elements_stability
        plt_stab = plot_stability_diagram(pd,elements_stability)
        if args.savefig:
            fname = 'SD_' + '-'.join(elements) + '.pdf'
            path = os.path.join(os.getcwd(),fname)
            plt_stab.savefig(path,bbox_inches='tight')
        else:
            plt_stab.show()
    
    if args.savejson:
        fname = 'PD_' + '-'.join(elements) + '.json'
        path = os.path.join(os.getcwd(),fname)
        save_object_as_json(pd,path)
    
    return


def get_chempots(pd,composition):
    return PDHandler(pd).get_all_boundaries_chempots(composition)
    
def get_phase_diagram(elements):  
    from pymatgen.ext.matproj import MPRester
    with MPRester(SETTINGS['API_KEY']) as mpr:
        compat = MaterialsProjectCompatibility()  # sets energy corrections and +U/pseudopotential choice
        unprocessed_entries = mpr.get_entries_in_chemsys(elements,inc_structure=True)
    processed_entries = compat.process_entries(unprocessed_entries)  # filter and add energy corrections    
    return PhaseDiagram(processed_entries)

def plot_pd(pd):
    return PDHandler(pd).get_plot()

def plot_stability_diagram(pd,axis_elements):
    return PDHandler(pd).get_stability_diagram(axis_elements)


   


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 14 12:38:27 2021

@author: lorenzo
"""
import matplotlib 
import matplotlib.pyplot as plt

def update_fontsize(size):
    matplotlib.rcParams.update({'font.size': size})
    return







def format_composition(string,all_math=False):
    
    def write(newc,new_string,skip,all_math):
        if not skip:
            if all_math:
                    new_string.append('_{%s}'%newc)
            else:
                new_string.append('$_{%s}$'%newc)
        return new_string
            
    skip = False
    new_string = []
    if all_math:
        new_string.append('$')
    sp = list(string)
    
    for c in sp:
        
        if c.isdigit():
            try:
                nextc = sp[sp.index(c)+1]
            except:
                nextc = ''
            if nextc.isdigit():
                newc = c + nextc
                new_string = write(newc,new_string,skip,all_math)
                skip = True
            else:
                newc = c
                new_string = write(newc,new_string,skip,all_math)
                skip = False

        else:
            new_string.append(c)
    if all_math:
        new_string.append('$')       
    new_string = ''.join(new_string)
    
    return new_string


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  5 16:33:14 2023

@author: villa
"""

import os
import os.path as op
import time
from datetime import datetime
import subprocess
import schedule
from shutil import which

from monty.dev import requires

from pynter import SETTINGS, run_local
from pynter.slurm.job_settings import JobSettings
from pynter.data.jobs import get_job_from_directory

def setup_hpc(subparsers):
    
    parser_hpc = subparsers.add_parser('HPC',help='Interface with High Performance Computer')
    
    subparsers_hpc = parser_hpc.add_subparsers()
    parser_sync = subparsers_hpc.add_parser('sync',help='Sync data with HPC with rsync')
    setup_sync(parser_sync)
    
    parser_job = subparsers_hpc.add_parser('job',help='Sync and run job on HPC')
    setup_job(parser_job)
    
    return


def setup_sync(parser):
    hostname = SETTINGS['HPC']['hostname']
    workdir = SETTINGS['HPC']['workdir']
    localdir = SETTINGS['HPC']['localdir']
    
    parser.add_argument('-dry','--dry-run',action='store_true',help='Dry run with rsync (default: %(default)s)',default=False,dest='dry_run')
    parser.add_argument('-exc','--exclude',action='append',help='Files to exclude from sync (default: %(default)s)',
                        default=['core.*','WAVECAR'],metavar='',type=str,dest='exclude')
    parser.add_argument('-host','--hostname',help='ssh alias of the HPC (default: %(default)s)',default=hostname,metavar='',type=str,dest='hostname')
    parser.add_argument('-ldir','--localdir',help='Local directory (default: %(default)s)',default=localdir,metavar='',type=str,dest='localdir')
    parser.add_argument('-wdir','--workdir',help='Remote directory (default: %(default)s)',default=workdir,metavar='',type=str,dest='workdir')
    parser.add_argument('-per','--periodic',action='store_true',help='Sync every hour with HPC (default: %(default)s)',default=False,dest='periodic')
    parser.set_defaults(func=run_sync)
    return



@requires(which("rsync"),
      "rsync needs to be installed, you can install it with 'sudo apt-get install rsync'.")
def run_sync(args):
    command = get_command(args)
    run_command(command,periodic=args.periodic)
    return

    
def get_command(args):
    workdir = op.join(args.workdir,'')  #ensure backslash at the end
    localdir = op.join(args.localdir,'')
    hostname = args.hostname
    
    cmd = "rsync -r -uavzh " #keep the spaces
    if args.dry_run:
        cmd += "--dry-run "
    if args.exclude:
        for s in args.exclude:
            cmd += f'--exclude={s} ' 
    cmd += f"-e ssh {hostname}:{workdir} {localdir} "   
    return cmd
    
    
def run_command(command,periodic=False):
    if periodic:
        homedir = os.getenv("HOME")
        wdir = os.path.join(homedir,'.pynter','.sync_hpc')
        if not os.path.exists(wdir):
            os.makedirs(wdir)
        os.chdir(wdir)
        command = command.split()
        
        def job(command):        		
        	proc = subprocess.run(command, capture_output=True, shell=False, text=True)
        	stdout = proc.stdout
        	stderr = proc.stderr 
        	dtn = datetime.now()
        	date = dtn.date().isoformat()
        	hour = dtn.ctime().split()[3]
        	path = os.path.join(wdir,date)
        	if not os.path.exists(path):
        		os.mkdir(path)
        	
        	outfile = os.path.join(path,'sync_'+ hour +'.out')
        	errfile = os.path.join(path,'sync_'+ hour +'.err')
        
        	with open(outfile,'w') as outf, open(errfile,'w') as errf:
        		outf.write(stdout)
        		errf.write(stderr)
        
        schedule.every(60).minutes.do(job,command)
        
        while True:
        	schedule.run_pending()
        	time.sleep(30)
        
    else:
        print('Starting rsync...')
        run_local(command,printout=True)
        return
        


def setup_job(parser):
    jsf = JobSettings().filename
    parser.add_argument('-p','--path',help='Job path (default: %(default)s)',required=False,type=str,
                        default=os.getcwd(),metavar='',dest='path')
    parser.add_argument('-dry','--dry-run',action='store_true',help='Dry run with rsync (default: %(default)s)',default=False,dest='dry_run')
    parser.add_argument('-st','--sync-to-hpc',help='Sync to hpc',required=False,default=False,action='store_true',dest='sync_to_hpc')
    parser.add_argument('-sf','--sync-from-hpc',help='Sync from hpc',required=False,default=False,action='store_true',dest='sync_from_hpc')
    parser.add_argument('-f','--filename',help='Name of bash job script (default: %(default)s)',required=False,default=jsf,type=str,metavar='',dest='filename')
    parser.add_argument('-R','--run',help='Run job on HPC',required=False,default=False,action='store_true',dest='run')
    parser.set_defaults(func=run_job)
    return 
    
def run_job(args):
    j = get_job_from_directory(path=args.path,job_script_filename=args.filename,load_outputs=False)
    
    if args.sync_from_hpc and args.sync_to_hpc:
        raise ValueError("Cannot sync to and from HPC simultaneously")
    elif args.sync_from_hpc:
        j.sync_from_hpc(dry_run=args.dry_run)
    elif args.sync_to_hpc:
        j.sync_to_hpc(dry_run=args.dry_run)
        
    if args.run:
        j.run_job(write_input=False,sync=False)
    
    return
    
    








import yaml
import os
import subprocess
import warnings

def get_config_from_default_file():
    """
    Get initial settings from config_default.yml
    """
    path = os.path.abspath(__file__).strip(os.path.basename(__file__))
    filename = 'config_default.yml'
    with open(os.path.join(path,filename),'r') as ymlfile:
        return yaml.load(ymlfile,Loader=yaml.FullLoader)
    
def get_vasp_defaults_from_default_file():
    """
    Get initial vasp settings from ./vasp/vasp_default.yml
    """
    from pynter import vasp
    path = os.path.abspath(vasp.__file__).strip(os.path.basename(vasp.__file__))
    filename = 'vasp_default.yml'
    with open(os.path.join(path,filename),'r') as ymlfile:
        return yaml.load(ymlfile,Loader=yaml.FullLoader)


SETTINGS = get_config_from_default_file()
SETTINGS['vasp'] = get_vasp_defaults_from_default_file()


def get_cfgfile():
    homedir = os.getenv("HOME")
    cfgfile = os.path.join(homedir,'.pynter','config.yml')
    return cfgfile

cfgfile = get_cfgfile()
def load_config(cfgfile=cfgfile):
    """
    Load dictionary with configuration from yaml cofiguration file. The default is in ~/.pynter/config.yml.

    Parameters
    ----------
    cfgfile : (str), optional
        Path to configuration file. The default is ~/.pynter/config.yml.

    Returns
    -------
    (dict)
        Configuration dictionary.
    """
    if os.path.exists(cfgfile):
        with open(cfgfile,"r") as ymlfile:
            return yaml.load(ymlfile,Loader=yaml.FullLoader) # add Loader to not get warning
    else:
        warnings.warn('%s does not exist. Run "pynter configure" in the terminal to create it. Using default settings for now. '%cfgfile)
        return


def get_vasp_cfgfile():
    homedir = os.getenv("HOME")
    cfgfile = os.path.join(homedir,'.pynter','vasp.yml')
    return cfgfile


cfgfile = get_vasp_cfgfile()
def load_vasp_default(cfgfile=cfgfile):
    """
    Load dictionary with VASP configuration from yaml cofiguration file. The default is in ~/.pynter/vasp.yml.

    Parameters
    ----------
    cfgfile : (str), optional
        Path to configuration file. The default is ~/.pynter/vasp.yml.

    Returns
    -------
    (dict)
        Configuration dictionary.
    """
    if os.path.exists(cfgfile):
        with open(cfgfile,"r") as ymlfile:
            return yaml.load(ymlfile,Loader=yaml.FullLoader) # add Loader to not get warning
    else:
        warnings.warn('%s does not exist. Run "pynter configure" in the terminal to create it. Using default settings for now'%cfgfile)
        return  

# Overwrite initial settings
config = load_config()
if config:
    SETTINGS.update(config)
    
vasp_defaults = load_vasp_default()
if vasp_defaults:
    SETTINGS['vasp'].update(vasp_defaults)



def run_local(cmd,printout=True,dry_run=False,**kwargs):
    """
    Run a command locally with subprocess package.

    Parameters
    ----------
    cmd : (str)
        Command to run.
    printout : (bool), optional
        Print command, output and error. The default is True.
    dry_run : (bool), optional
        Return back the command, without executing it. The default is False.
    **kwargs : (dict)
        Kwargs to pass to subprocess.run().

    Returns
    -------
    stdout
    stderr
    """
    if dry_run:
        if printout:
            print(cmd)
        return cmd, ''
    if printout:
        print("Run command: %s" %cmd)
    command = cmd.split()
    proc = subprocess.run(command, capture_output=True, shell=False, text=True,**kwargs)
    stdout = proc.stdout
    stderr = proc.stderr
    if printout:
        print(stdout)
        if stderr:
            print(stderr)
    return stdout, stderr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 12 15:27:33 2020

@author: villa
"""

from pymatgen.core.periodic_table import Element

def convert_conc_from_weight_to_cm3(c_weight,target_el,composition,bulk_structure):
    """
    Convert concentration of dopand from weight % (common experimental data) to cm^-3.

    Parameters
    ----------
    c_weight : (float)
        Defect concentration in weight %.
    target_el : (str)
        Symbol of target element.
    composition : (Composition)
        Composition (Pymatgen) of the material, uses only list of elements.
    bulk_structure : (Structure)
        Structure (Pymatgen) of bulk, needed for lattice volume.

    Returns
    -------
    conc : (float)
        Concentration in cm-3^.

    """
    sum_MM = sum([el.atomic_mass for el in composition.elements])
    r_MM = Element(target_el).atomic_mass / sum_MM
    conc = c_weight/r_MM * 1/bulk_structure.lattice.volume * 1e24 *0.01
    return conc




     



from abc import abstractmethod
import os
import os.path as op
import shutil

from pynter.slurm.job_settings import JobSettings
from pynter.slurm.interface import HPCInterface
from pynter.tools.utils import grep_list


class Job:
    
    def __init__(self,path=None,inputs=None,job_settings=None,outputs=None,job_script_filename=None,name=None):
        """
        Class to control and organize inputs and outputs of a generic job.

        Parameters
        ----------
        path : (str), optional
            Path where job is stored. The default is None. If None the work dir is used.
        inputs : (dict), optional
            Dictionary with input data. The default is None.
        job_settings : (dict), optional
            Dictionary with job settings. The default is None. Documentation in JobSettings class in slurm.job_script module
        outputs : (dict), optional
            Dictionary with output data. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is taken from the key 'filename' in the job_settings in the config file.
        name : (str)
            Name of the job. If none the name is searched in the job script.

        """
                
        self.path = path if path else os.getcwd()
        self.inputs = inputs
        self.job_settings = JobSettings(**job_settings) if job_settings else JobSettings()
        self.outputs = outputs
        self.job_script_filename = job_script_filename if job_script_filename else JobSettings().filename
        
        self._localdir = HPCInterface().localdir
        self._workdir = HPCInterface().workdir
        self.path_relative = op.abspath(self.path).replace(self._localdir,'')
        
        self.path_in_hpc = self._workdir + self.path_relative
        
        
        if outputs:
            self.get_output_properties()
        
        
        if name:
            self.name = name
        elif self.job_settings:
            self.name = self.job_settings['slurm']['job-name']
        elif op.isfile(op.join(self.path,self.job_script_filename)):
            s = JobSettings.from_bash_file(self.path,filename=self.job_script_filename)
            self.name = s.settings['slurm']['job-name']
        else:
            self.name = 'no_name'
            
        if not self.job_settings:
            self.job_settings = {}
        self.job_settings['slurm']['job-name'] = self.name


    def __str__(self):
        jobclass = self.jobclass
        if hasattr(self,'group'):
            if self.group != '':
                printout = '%s "%s" of group "%s"' %(jobclass, self.name, self.group)
            else:
                printout = '%s "%s"' %(jobclass, self.name)
        else:
            self.group = ''
            printout = '%s "%s"' %(jobclass, self.name)
        
        return printout
    
    def __repr__(self):
        return self.__str__()
        
    @property
    def jobclass(self):
        return self.__class__.__name__

        
    def cancel_job(self):
        """Cancel job on HPC"""
        hpc = HPCInterface()
        job_id = self.job_id()
        hpc.cancel_jobs(job_id)
        
        return 


    def copy(self):
        """
        Copy Job object
        """
        path=self.path
        inputs=self.inputs.copy() if self.inputs else None
        job_settings=self.job_settings.copy() if self.job_settings else None
        outputs=self.outputs.copy() if self.outputs else None
        job_script_filename=self.job_script_filename
        name=self.name
        return self.__class__(path=path,inputs=inputs,job_settings=job_settings,
                       outputs=outputs,job_script_filename=job_script_filename,
                       name=name)
    
    
    def delete_job_files(self,safety=True):
        """
        Delete Job folder (self.path)

        Parameters
        ----------
        safety : (bool), optional
            Ask confirmation to delete job. The default is True.
        """
        if safety:
            inp = input('Are you sure you want to delete Job %s ? (y/n) : ' %self.name)
            if inp in ('y','Y'):
                shutil.rmtree(self.path)
                print('Deleted Job %s'%self.name)
            else:
                print('Job %s is left unchanged'%self.name)
        else:
            shutil.rmtree(self.path)
            print('Deleted Job %s'%self.name)
        return


    @abstractmethod
    def get_inputs(self):
        pass

    @abstractmethod
    def get_outputs(self):
        pass
    
    @abstractmethod
    def get_output_properties(self):
        pass

    @abstractmethod
    def insert_in_database(self):
        pass


    def job_id(self):
        """Get job ID from the queue on HPC"""        
        hpc = HPCInterface()
        stdout,stderr = hpc.qstat(printout=False)
        queue = stdout.splitlines()
        job_lines = grep_list(self.name,queue)
        if job_lines == []:
            raise ValueError (f'Job named "{self.name}" is not currently running or pending')
        elif len(job_lines) > 1:
            raise ValueError (f'More than one job named "{self.name}" has been found in queue:\n{stdout}')
        else:
            job_line = job_lines[0].split()
            job_id = job_line[0]
        
        return job_id


    def job_queue(self):
        """
        Print job queue from HPC on screen
        
        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        hpc = HPCInterface()
        stdout,stderr = hpc.qstat()
        
        return stdout,stderr
        

    def run_job(self,write_input=True,sync=True):
        """
        Run job on HPC. Input files are automatically written and sync to HPC is performed.

        Parameters
        ----------
        write_input : (bool), optional
            Write input file stored in "inputs" dictionary. The default is True.
        sync : (bool), optional
            Sync files to HPC before running. The default is True

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        if write_input:
            self.write_input()
        hpc = HPCInterface()
        if sync:
            self.sync_to_hpc()
        stdout,stderr = hpc.sbatch(path=self.path_in_hpc,job_script_filename=self.job_script_filename)
        
        return stdout,stderr
    

    def sync_from_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync job data from HPC to local machine

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.

        """
        hpc = HPCInterface()
        abs_path = op.abspath(self.path)
        localdir = abs_path 
        stdout,stderr = hpc.rsync_from_hpc(remotedir=self.path_in_hpc,localdir=localdir,exclude=exclude,dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return
        
        
    def sync_to_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync job data from local machine to HPC

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.

        """
        hpc = HPCInterface()
        abs_path = op.abspath(self.path)
        localdir = abs_path 
        stdout,stderr = hpc.rsync_to_hpc(localdir=localdir,remotedir=self.path_in_hpc,exclude=exclude,dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return


    def get_status_from_queue(self,stdout):     
        """
        Get Job status from stdout of qstat command in HPC.

        Returns
        -------
        status : (str)
            Job status. Possible status are 'PENDING','RUNNING','NOT IN QUEUE'.
        """
        queue = stdout.splitlines()
        job_lines = grep_list(self.name,queue)
        if job_lines == []:
            status = 'NOT IN QUEUE'
        elif len(job_lines) > 1:
            raise ValueError (f'More than one job named "{self.name}" has been found in queue:\n{stdout}')
        else:
            job_line = job_lines[0].split()
            status = job_line[4]
            if status == 'PD':
                status = 'PENDING'
            if status == 'R':
                status = 'RUNNING'
            if status == 'CG':
                status = 'COMPLETED'
            
        return status


    def status(self):
        """
        Get job status from HPC. Gets queue from HPC and gets Job status from the queue.

        Returns
        -------
        status : (str)
            Job status. Possible status are 'PENDING','RUNNING','NOT IN QUEUE'.
        """
        hpc = HPCInterface()
        stdout,stderr = hpc.qstat(printout=False)
        status = self.get_status_from_queue(stdout)            
        return status
 
           
    @abstractmethod
    def write_input():
        pass
        


def get_job_from_directory(path=None,job_script_filename=None,load_outputs=True,jobs_kwargs=None):   
    """
    Get Job object froma directory. The job type is selected based on the content of the folder.
    - If ("INCAR","KPOINTS","POSCAR","POTCAR") files are present, VaspJob is initialized.
    - If ("INCAR","KPOINTS","POTCAR") files are present and "POSCAR" is not present, VaspNEBJob is initialized.

    Parameters
    ----------
    path : (str)
        Directory containing job.
    job_script_filename : (str), optional
        Filename of job script. The default is taken from the key 'filename' in the job_settings in the config file.
    sort : (str or list), optional
        Sort list of jobs by features. If False or None jobs are not sorted. The default is 'name'.
    load_outputs : (bool)
        Load job outputs. The default is True.
    jobs_kwargs : (dict), optional
        Dictionary with job class name as keys and kwargs as values. Kwargs to be used when importing job 
        from directory for each job class.

    Returns
    -------
    jobs : (list)
        List of Job objects.
    """    
    from pynter.vasp.jobs import VaspJob, VaspNEBJob
    
    path = path if path else os.getcwd()
    job_script_filename = job_script_filename if job_script_filename else JobSettings().filename
        
    for root,dirs,files in os.walk(path):      
        if all(f in files for f in ['INCAR','KPOINTS','POSCAR','POTCAR']):
            path = op.abspath(root)
            if jobs_kwargs:
                kwargs = jobs_kwargs['VaspJob'] if 'VaspJob' in jobs_kwargs.keys() else {}
            else:
                kwargs = {}
            j = VaspJob.from_directory(path,job_script_filename=job_script_filename,load_outputs=load_outputs,**kwargs)
            return j
        elif all(f in files for f in ['INCAR','KPOINTS','POTCAR']) and 'POSCAR' not in files:
            path = op.abspath(root)
            j = VaspNEBJob.from_directory(path,job_script_filename=job_script_filename,load_outputs=load_outputs)
            return j


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr  3 14:08:42 2023

@author: villa
"""
from abc import ABCMeta, abstractmethod, abstractproperty
from monty.json import MSONable
from pymatgen.core.composition import Composition
from pymatgen.core.sites import PeriodicSite
import importlib
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from pynter.tools.structure import is_site_in_structure, is_site_in_structure_coords, remove_oxidation_state_from_site
# Adapted from pymatgen.analysis.defect.core 


class Defect(MSONable,metaclass=ABCMeta): #MSONable contains as_dict and from_dict methods
    """
    Abstract class for a single point defect
    """

    def __init__(self, defect_site, bulk_structure, charge=None, multiplicity=None,label=None):
        """
        Base class for defect objets

        Parameters
        ----------
        defect_site : (Site)
            Pymatgen Site object of the defect.
        bulk_structure : (Structure)
            Pymatgen Structure without defects.
        charge : (int or float)
            Charge of the defect.
        multiplicity : (int)
            Multiplicity of defect within the simulation cell.
        """
        self._defect_site = defect_site
        self._bulk_structure = bulk_structure
        self._charge = charge
        self._multiplicity = multiplicity 
        self._label = label


    def __repr__(self):
        return "{} : {} {}".format(self.defect_type,self.site.frac_coords, self.name.dspecie)
    
    def __print__(self):
        return self.__repr__()

    @property
    def bulk_structure(self):
        """
        Structure without defects.
        """
        return self._bulk_structure
    
    @property
    def charge(self):
        """
        Charge of the defect.
        """
        return self._charge
     
    @abstractproperty
    def defect_composition(self):
        """
        Defect composition as a Composition object
        """
        return

    @abstractproperty
    def defect_site_index(self):
        """
        Index of the defect site in the structure
        """
        return
    
    @property
    def defect_specie(self):
        return self.name.dspecie
    
    @property
    def defect_structure(self):
        """
        Structure of the defect
        """
        return self.generate_defect_structure() 
    
    @property
    def defect_type(self):
        return self.__class__.__name__
    
    @property
    def delta_atoms(self):
        """
        Dictionary with defect element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return

    @abstractmethod
    def get_multiplicity(self):
        return 

    @property
    def label(self):
        return self._label

    @property
    def multiplicity(self):
        """
        Multiplicity of a defect site within the structure
        """
        return self._multiplicity
      
    @abstractproperty  
    def name(self):
        """
        Name of the defect
        """
        return

    @property
    def site(self):
        """
        Defect position as a site object
        """
        return self._defect_site
   
    @property
    def site_concentration_in_cm3(self):
        return self.multiplicity * 1e24 / self.bulk_structure.volume 
   
    @property  
    def symbol(self):
        """
        Latex formatted name of the defect
        """
        return self.name.symbol

    @property
    def symbol_with_charge(self):
        """
        Name in latex format with charge written as a number.
        """
        return format_legend_with_charge_number(self.symbol,self.charge)
    
    @property
    def symbol_with_charge_kv(self):
        """
        Name in latex format with charge written with kroger and vink notation.
        """
        return format_legend_with_charge_kv(self.symbol,self.charge)

    def set_charge(self, new_charge=0.0):
        """
        Sets the charge of the defect.
        """
        self._charge = new_charge
        return
    
    def set_label(self,new_label):
        """
        Sets the label of the defect
        """
        self._label = new_label
        return

    def set_multiplicity(self, new_multiplicity=1):
        """
        Sets the charge of the defect.
        """
        self._multiplicity = new_multiplicity
        return
        
        
class Vacancy(Defect):
    
    """
    Subclass of Defect for single vacancies.
    """
    @property
    def defect_composition(self):
        """
        Composition of the defect.
        """
        temp_comp = self.bulk_structure.composition.as_dict()
        temp_comp[str(self.site.specie)] -= 1
        return Composition(temp_comp)
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the bulk structure
        """
        _,index = is_site_in_structure_coords(self.site, self.bulk_structure)
        return index
    
    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.site.specie.symbol:-1}
    
    def generate_defect_structure(self,bulk_structure=None,defect_site_index=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : (Structure), optional
            Bulk Structure. If not provided self.bulk_structure is used. The default is None.
        defect_site_index : (Structure), optional
            Index of the defect site in the bulk structure. If not provided 
            self.defect_site_index is used. The default is None.

        Returns
        -------
        structure : (Structure)
            Structure containing the defect.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_site_index = defect_site_index if defect_site_index else self.defect_site_index
        defect_structure = bulk_structure.copy()
        defect_structure.remove_sites([defect_site_index])
        return defect_structure
    
    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure

        Parameters
        ----------
        **kwargs : (dict)
            Arguments to pass to SpacegroupAnalyzer (symprec, angle_tolerance)
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site)
        return len(equivalent_sites)

    @property
    def name(self):
        """
        Name of the defect. Behaves like a string with additional attributes.
        """
        return DefectName(self.defect_type,self.site.specie.symbol,None,self.label)
        
    
class Substitution(Defect):
    """
    Subclass of Defect for substitutional defects.
    """

    def __init__(self,defect_site,bulk_structure,charge=None,multiplicity=None,label=None,site_in_bulk=None):
        """
        site_in_bulk: (PeriodicSite)
            Original Site in bulk structure were substitution took place
        """
        super().__init__(defect_site,bulk_structure,charge,multiplicity,label)  
        self._site_in_bulk = site_in_bulk if site_in_bulk else self.get_site_in_bulk()


    @property  
    def defect_composition(self):
        """
        Composition of the defect.
        """
        defect_index = self.defect_site_index
        temp_comp = self.bulk_structure.composition.as_dict()
        temp_comp[str(self.site.specie)] += 1
        temp_comp[str(self.bulk_structure[defect_index].specie)] -= 1
        return Composition(temp_comp)

    @property
    def defect_site_index(self):
        """
        Index of the defect site in the bulk structure
        """
        return self.bulk_structure.index(self.site_in_bulk)

    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.site.specie.symbol:1, self.site_in_bulk.specie.symbol:-1}
    
    def generate_defect_structure(self,bulk_structure=None,defect_site_index=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : (Structure), optional
            Bulk Structure. If not provided self.bulk_structure is used. The default is None.
        defect_site_index : (Structure), optional
            Index of the defect site in the bulk structure. If not provided 
            self.defect_site_index is used. The default is None.

        Returns
        -------
        structure : (Structure)
            Structure containing the defect.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_site_index = defect_site_index if defect_site_index else self.defect_site_index
        defect_structure = bulk_structure.copy()
        defect_structure.replace(defect_site_index,self.defect_specie)  
        return defect_structure

    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure

        Parameters
        ----------
        **kwargs : (dict)
            Arguments to pass to SpacegroupAnalyzer (symprec, angle_tolerance)
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site_in_bulk)
        return len(equivalent_sites)

    @property 
    def name(self):
        """
        Name for this defect. Behaves like a string with additional attributes.
        """
        return DefectName(self.defect_type,self.site.specie.symbol,self.site_in_bulk.specie.symbol,self.label)   
    
    @property
    def site_in_bulk(self):
        return self._site_in_bulk
    
    def get_site_in_bulk(self):
        try:
            site = min(
                self.bulk_structure.get_sites_in_sphere(self.site.coords, 0.5, include_index=True),
                           key=lambda x: x[1])  
            # there's a bug in pymatgen PeriodicNeighbour.from_dict and the specie attribute, get PeriodicSite instead
            site = PeriodicSite(site.species, site.frac_coords, site.lattice)
            return site
        except:
            return ValueError("""No equivalent site has been found in bulk, defect and bulk structures are too different.\
Try using the unrelaxed defect structure or provide bulk site manually""")
    
 
    
class Interstitial(Defect):
    """
    Subclass of Defect for interstitial defects.
    """
    
    @property
    def defect_composition(self):
        """
        Composition of the defect.
        """
        temp_comp = self.bulk_structure.composition.as_dict()
        temp_comp[str(self.site.specie)] += 1
        return Composition(temp_comp)
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the defect structure
        """
        is_site,index = is_site_in_structure(self.site, self.defect_structure)
        return index  # more flexibility with is_site_in_structure
       
    @property
    def delta_atoms(self):
        """
        Dictionary with element symbol as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {self.site.specie.symbol:1}
    
    def generate_defect_structure(self,bulk_structure=None):
        """
        Generate a structure containing the defect starting from a bulk structure.

        Parameters
        ----------
        bulk_structure : (Structure), optional
            Bulk Structure. If not provided self.bulk_structure is used. The default is None.

        Returns
        -------
        structure : (Structure)
            Structure containing the defect.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        defect_structure = bulk_structure.copy()
        defect_structure.append(self.site.species,self.site.frac_coords)
        return defect_structure
        

    def get_multiplicity(self):
        raise NotImplementedError('Not implemented for Interstitial')

    @property
    def name(self):
        """
        Name of the defect. Behaves like a string with additional attributes.
        """
        return DefectName(self.defect_type,self.site.specie.symbol,None,self.label)
        
      
class Polaron(Defect):
    """
    Subclass of Defect for polarons
    """   
    
    def __init__(self,defect_site,bulk_structure,charge=None,multiplicity=None,
                 label=None,defect_structure=None):
        """
        defect_structure: (Structure)
            Structure containing the polaron. If not provided the site index is searched 
            in the bulk structure, and the defect_structure is set equal to the bulk structure.
        """
        super().__init__(defect_site,bulk_structure,charge,multiplicity,label) 
        self._defect_structure = defect_structure
        
        
    @property
    def defect_composition(self):
        """
        Composition of the defect
        """
        return self.bulk_structure.composition
    
    @property
    def defect_site_index(self):
        """
        Index of the defect site in the structure
        """
        return self.defect_structure.index(self.site)
        
    @property
    def delta_atoms(self):
        """
        Dictionary with delement as keys and difference in particle number 
        between defect and bulk structure as values
        """
        return {}
    
    def generate_defect_structure(self,bulk_structure=None):
        """
        Structure containing the polaron. If not provided the site index is searched 
        in the bulk structure, and the defect_structure is set equal to the bulk structure.
        """
        if self._defect_structure:
            return self._defect_structure
        else:
            bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
            return bulk_structure  
        
        
    def get_multiplicity(self,**kwargs):
        """
        Get multiplicity of the defect in the structure

        Parameters
        ----------
        **kwargs : (dict)
            Arguments to pass to SpacegroupAnalyzer (symprec, angle_tolerance)
        """
        sga = SpacegroupAnalyzer(self.bulk_structure,**kwargs)
        symmetrized_structure = sga.get_symmetrized_structure()
        equivalent_sites = symmetrized_structure.find_equivalent_sites(self.site)
        return len(equivalent_sites)

    @property
    def name(self):
        """
        Name of the defect. Behaves like a string with additional attributes.
        """
        return DefectName(self.defect_type,self.site.specie.symbol,None,self.label)
  
    
    
class DefectComplex(MSONable,metaclass=ABCMeta):

    def __init__(self,defects,bulk_structure,charge=None,multiplicity=None,label=None):
        """
        Class to describe defect complexes

        Parameters
        ----------
        defects : (list)
            List of Defect objects.
        bulk_structure : (Structure)
            Pymatgen Structure of the bulk material.
        charge : (int or float)
            Charge of the defect.
        multiplicity : (int)
            Multiplicity of the defect.
        """
        self._defects = defects
        self._bulk_structure = bulk_structure
        self._charge = charge
        self._multiplicity = multiplicity
        self._label = label


    def __repr__(self):
        return "{} : {}".format(self.defect_type,
        [(d.defect_type,d.name.dspecie,d.site.frac_coords.__str__()) 
         for d in self.defects])

    def __str__(self):
        return self.__repr__()
    
    @property
    def bulk_structure(self):
        """
        Returns the structure without any defects.
        """
        return self._bulk_structure
    
    @property
    def charge(self):
        """
        Charge of the defect.
        """
        return self._charge

    @property
    def defects(self):
        """
        List of single defects consituting the complex.
        """
        return self._defects

    @property
    def defect_composition(self):
        return self.defect_structure.composition
        
    @property
    def defect_names(self):
        return [d.name for d in self.defects]

    @property
    def defect_structure(self):
        return self.generate_defect_structure()

    @property
    def defect_type(self):
        return self.__class__.__name__
    
    @property
    def delta_atoms(self):
        """
        Dictionary with Element as keys and particle difference between defect structure
        and bulk structure as values.
        """
        da_global = None
        for d in self.defects:
            da_single = d.delta_atoms
            if da_global is None:
                da_global = da_single.copy()
            else:
                for e in da_single:
                    prec = da_global[e] if e in da_global.keys() else 0
                    da_global[e] = prec + da_single[e]       
        return da_global 

    def generate_defect_structure(self,bulk_structure=None):
        """
        Generate a structure containing the defect starting from a bulk structure.
        If not provided self.bulk_structure is used.
        """
        bulk_structure = bulk_structure if bulk_structure else self.bulk_structure
        structure = bulk_structure.copy()
        for df in self.defects:
            df_structure = df.generate_defect_structure(structure)
            structure = df_structure.copy()
        return structure

    def get_multiplicity(self):
        raise NotImplementedError('Not implemented for DefectComplex')

    @property
    def label(self):
        return self._label

    @property
    def multiplicity(self):
        """
        Multiplicity of a defect site within the structure
        """
        return self._multiplicity

    @property
    def name(self):
        """
        Name of the defect. Behaves like a string with additional attributes.
        """
        return DefectComplexName([d.name for d in self.defects],self.label)

    @property
    def site_concentration_in_cm3(self):
        return self.multiplicity * 1e24 / self.bulk_structure.volume 
    
    @property
    def sites(self):
        return [d.site for d in self.defects]
    
    @property
    def symbol(self):
        return self.name.symbol

    @property
    def symbol_with_charge(self):
        """
        Name in latex format with charge written as a number.
        """
        return format_legend_with_charge_number(self.symbol,self.charge)
    
    @property
    def symbol_with_charge_kv(self):
        """
        Name in latex format with charge written with kroger and vink notation.
        """
        return format_legend_with_charge_kv(self.symbol,self.charge)


    def set_charge(self, new_charge):
        """
        Sets the charge of the defect.
        """
        self._charge = new_charge
        return    

    def set_label(self,new_label):
        """
        Sets the label of the defect
        """
        self._label = new_label
        return

    def set_multiplicity(self, new_multiplicity):
        """
        Sets the charge of the defect.
        """
        self._multiplicity = new_multiplicity
        return
    

class DefectName(str,MSONable):
    
    def __new__(cls, dtype,dspecie,bulk_specie=None,label=None):
        if dtype == 'Vacancy':
            name = 'Vac_%s' %dspecie
            symbol = '$V_{%s}$' %dspecie
            
        elif dtype == 'Substitution':
            name = 'Sub_%s_on_%s' %(dspecie,bulk_specie)
            symbol = '$%s_{%s}$' %(dspecie,bulk_specie)
            
        elif dtype == 'Interstitial':
            name = 'Int_%s' %dspecie
            symbol = '$%s_{i}$' %dspecie
            
        elif dtype == 'Polaron':
            name = 'Pol_%s' %dspecie
            symbol = '$%s_{%s}$' %(dspecie,dspecie)
            
        if label:
            fullname = name +'(%s)'%label
            symbol = symbol +'(%s)'%label
        else:
            fullname = name
        
        #get str instance
        instance = super().__new__(cls,fullname) 
        instance._dtype = dtype
        instance._dspecie = dspecie
        instance._bulk_specie = bulk_specie
        instance._name = name
        instance._label = label
        instance._fullname = fullname
        instance._symbol = symbol
        
        return instance


    def __init__(self,dtype,dspecie,bulk_specie=None,label=None):
        """
        Class to systematically organize the name of a defect.
        Behaves like a string, with the additional attributes.

        Parameters
        ----------
        dtype : (str)
            Type of the defect. Types are "Vacancy","Substitution","Interstitial","Polaron".
        dspecie : (str)
            Element of defect specie.
        bulk_specie : (str)
            Element of the specie that has been replaced. To be set only for "Substitution".
        label : (str), optional
            Label for the defect. Will be displayed in parenthesis after the name. The default is None.
        """
        super().__init__() # init string


    def __str__(self):
        return self.fullname.__str__()
    
    def __repr__(self):
        return self.fullname.__repr__()
    
    def __iter__(self):
        return [self].__iter__() # dummy iter to handle single defecs and complexes the same way 

    def __eq__(self, other):
        if isinstance(other, str):
            return self.fullname == other
        elif isinstance(other, DefectName):
            return self.fullname == other.fullname
        else:
            return False

    def __hash__(self):
        return hash(self.fullname)

    @property    
    def bulk_specie(self):
        if hasattr(self, '_bulk_specie'):
            bulk_specie = self._bulk_specie
        else:
            bulk_specie = None
        return bulk_specie

    @property        
    def dspecie(self):
        return self._dspecie

    @property
    def dtype(self):
        return self._dtype

    @property
    def fullname(self):
        return self._fullname

    @property
    def label(self):
        return self._label

    @property
    def name(self):
        return self._name

    @property
    def symbol(self):
        return self._symbol
    
    
    @staticmethod
    def from_string(string):
        args = DefectName._get_args_from_string(string)
        return DefectName(*args)
          
    def _get_args_from_string(string):
        if '(' in string:
            name,label = string.split('(')
            label = label.strip(')')
        else:
            name = string
            label=None
        nsplit = name.split('_')
        ntype = nsplit[0]
        el = nsplit[1]
        bulk_specie = None
        if ntype=='Vac':
            dtype = 'Vacancy'
            dspecie = el
        elif ntype=='Int':
            dtype = 'Interstitial'
            dspecie = el
        elif ntype=='Sub':
            dtype = 'Substitution'
            dspecie = el
            el_bulk = nsplit[3]
            bulk_specie = el_bulk
        elif ntype=='Pol':
            dtype = 'Polaron'
            dspecie = el
        return dtype,dspecie,bulk_specie,label
            
            
class DefectComplexName(str,MSONable):
    
    def __new__(cls, defect_names, label=None):
        name = '-'.join([n.name for n in defect_names]) #exclude labels
        if label:
            fullname =  name + '(%s)'%label
        else:
            fullname = name
        instance = super().__new__(cls,fullname)        
        instance._name = name
        instance._label = label
        instance._fullname = fullname
        instance._defect_names = defect_names
        return instance
    
    def __init__(self,defect_names,label=None):
        """
        Class to systematically organize the name of a defect complex. The name of the 
        complex concatenates with '-' the names of the single defects.
        Behaves like a string, with the additional attributes.

        Parameters
        ----------
        defect_names : (list)
            List of DefectName objects.
        label : (str), optional
            Label for the defect. Will be displayed in parenthesis after the name. The default is None.
        """
        super().__init__()
        
                
    def __str__(self):
        return self.fullname.__str__()
    
    def __repr__(self):
        return self.fullname.__repr__()
    
    def __iter__(self):
        return self.defect_names.__iter__()

    def __eq__(self, other):
        if isinstance(other, str):
            return self.fullname == other
        elif isinstance(other, DefectName):
            return self.fullname == other.fullname
        else:
            return False

    def __hash__(self):
        return hash(self.fullname)

    @property
    def defect_names(self):
        return self._defect_names

    @property
    def dspecies(self):
        return [n.dspecie for n in self.defect_names]
    
    @property
    def dtype(self):
        return 'DefectComplex'
        
    @property
    def fullname(self):
        return self._fullname

    @property
    def label(self):
        return self._label

    @property
    def name(self):
        return self._name
    
    @property
    def symbol(self):
        symbol = '-'.join([n.symbol.split('(')[0] for n in self.defect_names]) 
        if self.label:
            symbol = symbol + '(%s)'%self.label 
        return symbol
    
    
    @staticmethod
    def from_string(string):
        args = DefectComplexName._get_args_from_string(string)
        return DefectComplexName(*args)
    
    def _get_args_from_string(string):
        if '(' in string:
            name,label = string.split('(')
            label = label.strip(')')
        else:
            name = string
            label=None
        names = name.split('-')
        defect_names = [DefectName.from_string(n) for n in names]
        return defect_names,label


def get_defect_name_from_string(string):
    if '-' in string:
        return DefectComplexName.from_string(string)
    else:
        return DefectName.from_string(string)


def create_interstitials(structure,elements,supercell_size=None,**kwargs):
    """
    Create Interstitial objects based on Voronoi with pymatgen. 

    Parameters
    ----------
    structure : (Structure)
        Bulk structure.
    elements : (list)
        List of element symbols.
    supercell_size : (int), optional
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. The default is None.
    kwargs: 
        Arguments to pass to VoronoiInterstitialGenerator:
            clustering_tol: Tolerance for clustering the Voronoi nodes.
            min_dist: Minimum distance between an interstitial and the nearest atom.
            ltol: Tolerance for lattice matching.
            stol: Tolerance for structure matching.
            angle_tol: Angle tolerance for structure matching.
            kwargs: Additional keyword arguments for the ``TopographyAnalyzer`` constructor.

    Returns
    -------
    defects : (list)
        List of Interstitial objects
    """
    from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
    
    defects = []
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    generator = VoronoiInterstitialGenerator().generate(bulk_structure,elements)
    for inter in generator:
        bulk_structure.remove_oxidation_states()
        remove_oxidation_state_from_site(inter.site)
        interstitial = Interstitial(inter.site, bulk_structure,multiplicity=inter.multiplicity,
                                    label=f'mult{inter.multiplicity}')
        defects.append(interstitial)
    return defects


def create_substitutions(structure,elements_to_replace,supercell_size=None,site_indexes=None):
    """
    Create Substitution objects starting from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements_to_replace : (str), optional
        Dict with element symbol of specie to be replaced as keys and element 
        symbol of the species to be replaced with as values ({'old_El':'new_El'}).
    supercell_size : (int or numpy array), optional
        Input for the generate_defect_structure function of the old pymatgen Substitution class.
    site_indexes : (list), optional
        Site indexes where the defects need to be created. If None the first site found is 
        used. The default is None.

    Returns
    -------
    defects : (list)
        List of Substitution objects
    """
    defects = [] 
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    if site_indexes:
        sites = [bulk_structure[i] for i in site_indexes]
    else:
        sites = bulk_structure.sites
    for el_to_sub,el_subbed in elements_to_replace.items():
        for site in sites:
            if site.specie.symbol == el_to_sub:
                defect_site = PeriodicSite(el_subbed,site.frac_coords,site.lattice)
                defects.append(Substitution(defect_site, bulk_structure,site_in_bulk=site))
                if not site_indexes:
                    break
    return defects   


def create_vacancies(structure,elements=None,supercell_size=None,site_indexes=None):
    """
    Create structures with vacancies starting from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements : (str), optional
        Symbol of the elements for which vacancies are needed.
        If None all of the elements are considered. The default is None.
    supercell_size : (int or numpy array), optional
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. The default is None.
    site_indexes : (list), optional
        Site indexes where the defects need to be created. If None the first site found is 
        used. The default is None.

    Returns
    -------
    defects : (list)
        List of Vacancy objects
    """
    defects = []
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    if not elements:
        elements = [el.symbol for el in bulk_structure.composition.elements]
    if site_indexes:
        sites = [bulk_structure[i] for i in site_indexes]
    else:
        sites = bulk_structure.sites
    for el in bulk_structure.composition.elements:
        for site in sites:
            if el.symbol in elements:
                if site.specie == el:
                    vacancy = Vacancy(site, bulk_structure)
                    defects.append(vacancy)
                    if not site_indexes:
                        break
    return defects        


def format_legend_with_charge_number(label,charge):
    """
    Get label in latex format with charge written as a number.
   
    Parameters
    ----------
    label : (str)
        Original name of the defect.
    charge : (int or float)
        Charge of the defect. Floats are converted to integers.
    """
    s = label
    charge = int(charge)
    if charge > 0:
        q = '+'+str(charge)
    elif charge == 0:
        q = '\;' + str(charge)
    else:
        q = str(charge)
    return s + '$^{' + q + '}$'


def format_legend_with_charge_kv(label,charge):
    """
    Get label in latex format with charge written with kroger and vink notation.

    Parameters
    ----------
    label : (str)
        Original name of the defect.
    charge : (int or float)
        Charge of the defect. Floats are converted to integer.
    """
    mod_label = label + '$'
    charge = int(charge)
    if charge < 0:
        for i in range(0,abs(charge)):
            if i == 0:
                mod_label = mod_label + "^{"
            mod_label = mod_label + "'"
        mod_label = mod_label + "}"
    elif charge == 0:
        mod_label = mod_label + "^{x}"
    elif charge > 0:
        for i in range(0,charge):
            if i == 0:
                mod_label = mod_label + "^{"
            mod_label = mod_label + "°"
        mod_label = mod_label + "}"
    
    mod_label = mod_label + "$"
    
    return mod_label

    
def get_delta_atoms(structure_defect,structure_bulk):
    """
    Function to build delta_atoms dictionary starting from Pymatgen Structure objects.
    ----------
    structure_defect : (Pymatgen Structure object)
        Defect structure.
    structure_bulk : (Pymatgen Structure object)
        Bulk structure.
    Returns
    -------
    delta_atoms : (dict)
        Dictionary with Element as keys and delta n as values.
    """
    comp_defect = structure_defect.composition
    comp_bulk = structure_bulk.composition
        
    return get_delta_atoms_from_comp(comp_defect, comp_bulk)


def get_delta_atoms_from_comp(comp_defect,comp_bulk):
    """
    Function to biuld delta_atoms dictionary starting from Pymatgen Structure objects.
    ----------
    comp_defect : (Pymatgen Composition object)
        Defect structure.
    comp_bulk : (Pymatgen Composition object)
        Bulk structure.
    Returns
    -------
    delta_atoms : (dict)
        Dictionary with Element as keys and delta n as values.
    """
    delta_atoms = {}
    for el,n in comp_defect.items():
        nsites_defect = n
        nsites_bulk = comp_bulk[el] if el in comp_bulk.keys() else 0
        delta_n = nsites_defect - nsites_bulk
        if delta_n != 0:
            delta_atoms[el] = delta_n
        
    return delta_atoms    
    

def get_old_pmg_object(defect):
    """
    Convert defect object to the old version of pymatgen defect object.
    """
    module = importlib.import_module("pynter.defects.pmg.pmg_defects_core")
    pmg_class = getattr(module,defect.defect_type)
    charge = defect.charge if defect.charge else 0
    pmg_object = pmg_class(structure=defect.bulk_structure,defect_site=defect.site,
                           charge=charge,multiplicity=defect.multiplicity)
    return pmg_object

def get_pmg_object(defect):
    """
    Convert defect object to pymatgen defect object.
    """
    module = importlib.import_module("pymatgen.analysis.defects.core")
    pmg_class = getattr(module,defect.defect_type)
    pmg_object = pmg_class(structure=defect.bulk_structure,site=defect.site,
                           multiplicity=defect.multiplicity)
    return pmg_object
    




import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from pymatgen.core.periodic_table import Element
from pymatgen.analysis.eos import EOS


matplotlib.rcParams.update({'font.size': 15})


class DatasetAnalysis:
    
    def __init__(self,jobs):
        """
        Class to analyse multiple Jobs

        Parameters
        ----------
        jobs : 
            List of VaspJob objects
        """
        self.jobs = jobs


    def birch_murnaghan(self,title=None):
        """
        Fit E,V data with Birch-Murnaghan EOS

        Parameters
        ----------
        title : (str)
            Plot title

        Returns
        -------
        plt :
            Matplotlib object.
        eos_fit : 
            Pymatgen EosBase object.
        """
        V,E = [],[]
        for j in self.jobs:
            V.append(j.initial_structure.lattice.volume)
            E.append(j.final_energy)
        eos = EOS(eos_name='birch_murnaghan')
        eos_fit = eos.fit(V, E)
        eos_fit.plot(width = 10, height = 10 , text = '', markersize = 15,  label= 'Birch-Murnaghan fit')  
        plt.legend(loc=2, prop={'size': 20})
        if title:
            plt.title(title,size=25)
        plt.tight_layout()
        
        return plt, eos_fit


    def plot_convergence_encut(self,title=None,delta=1e-03,get_value=False):
        """
        Plot total energy convergence with respect to ENCUT

        Parameters
        ----------
        title : (str), optional
            Plot title. The default is None.
        delta : (float), optional
            Value of delta E for energy convergence. The default is 1e-03.
        get_value : (bool)
            Return also converged cutoff value.

        Returns
        -------
        plt : 
            Matplotlib object. Returns a tuple with converged cutoff value if get_value is True.
        """
        plt.figure(figsize=(10,8))
        cutoff_max = 0
        energies = {}
        for j in self.jobs:
            cutoff = j.incar['ENCUT']
            if cutoff > cutoff_max:
                cutoff_max = cutoff
            energies[cutoff] = j.final_energy/len(j.initial_structure)
        
        E0 = energies[cutoff_max]
        cutoffs = [c for c in energies.keys()]
        deltas = [E0 - e for e in energies.values()]
        
        convergence = False
        for i in range(0,len(cutoffs)):
            if i != 0:
                if i != len(cutoffs)-1:
                    if abs(deltas[i] - deltas[i-1]) < delta and abs(deltas[i] - deltas[i+1]) < delta:
                        convergence = True
                else:
                    if abs(deltas[i] - deltas[i-1]) < delta:
                        convergence = True       
                if convergence:
                    conv_cutoff = cutoffs[i]
                    conv_delta = deltas[i]
                    break
        
        if not convergence:
            print(f'Convergence not reached for delta E = {delta}')    

        plt.plot(cutoffs,deltas,'o--',markersize=10)
        if convergence:
            plt.scatter(conv_cutoff,conv_delta, color=[], edgecolor='r', linewidths=3, s=450)
        plt.xlabel('PW cutoff (eV)')
        plt.ylabel('Energy difference (eV/atom)')
        plt.grid()

        if get_value:
            return plt,conv_cutoff
        else:
            return plt         


    def plot_convergence_kpoint(self,title=None,delta=1e-03,get_value=False):
            """
            Plot total energy convergence with respect to K-Point Mesh.
            
            Parameters
            ----------
            title : (str), optional
                Plot title. The default is None.
            delta : (float), optional
                Value of delta E for energy convergence. The default is 1e-03.
            get_value : (bool)
                Return also converged cutoff value.
            Returns
            -------
            plt :
                Matplotlib object. Returns a tuple with converged cutoff value if get_value is True.
            """
            plt.figure(figsize=(10,8))
            kpoint_sum_max = 0
            energies = {}
            for j in self.jobs:
                kpoint = j.kpoints.kpts[0]
                kpoint_label = 'x'.join([str(k) for k in kpoint])
                kpoint_sum = len(j.computed_entry.data['actual_kpoints']) #count total number of kpoints
                if kpoint_sum > kpoint_sum_max:
                    kpoint_sum_max = kpoint_sum
                    kpoint_max = kpoint_label
                energies[kpoint_label] = j.final_energy/len(j.initial_structure)
    
            E0 = energies[kpoint_max]
            kpoints = [c for c in energies.keys()]
            deltas = [-E0 + e for e in energies.values()]
    
            convergence = False
            for i in range(0,len(kpoints)):
                if i != 0:
                    if i != len(kpoints)-1:
                        if abs(deltas[i] - deltas[i-1]) < delta and abs(deltas[i] - deltas[i+1]) < delta:
                            convergence = True
                    else:
                        if abs(deltas[i] - deltas[i-1]) < delta:
                            convergence = True
                    if convergence:
                        conv_kpoint = kpoints[i]
                        conv_delta = deltas[i]
                        break
    
            if not convergence:
                print(f'Convergence not reached for delta E = {delta}')
    
            plt.plot(kpoints,deltas,'o--',markersize=10)
            if convergence:
                plt.scatter(conv_kpoint,conv_delta, color=[], edgecolor='r', linewidths=3, s=450)
            plt.xlabel('K-Point Mesh')
            plt.ylabel('Energy difference (eV/atom)')
            plt.grid()

        
    def plot_fractional_charge(self,reference='electrons',name='',new_figure=True,legend_out=False):
        """
        Plot fractional charge study
        
        Parameters
        ----------
        reference: (str)
            Reference for fractional charge sign. 'electrons' or 'holes'. Default is 'electrons'.
        name: (str)
            Name to put in legend
        new_figure: (bool)
            If true initialize new matplotlib figure
        legend_out: (bool)
            If True place legend outside of the plot

        Returns
        -------
        plt : 
            Matplotlib object
        """
        if new_figure:
            plt.figure(figsize=(8,6))            
        jobs = self.jobs
        energy_dict = {}
        
        for j in jobs:
            if reference=='electrons':
                n = -1*np.around(j.charge,decimals=1) # express in terms of occupation
            if reference=='holes':
                n = 1 - np.around(j.charge,decimals=1) #shift to 0 charge of +1
            energy_dict[n] = j.final_energy            
        energy_dict = {k: v for k, v in sorted(energy_dict.items(), key=lambda item: item[0])} #order by charge

        def linear(x,E0,E1):
            return (E1-E0)*x + E0
        
        e_rescaled = {}
        e0 = float(energy_dict[0])
        e1 = float(energy_dict[1])
        for n in energy_dict:
            e = float(energy_dict[n])
            e_resc = e - linear(n, e0, e1)
            e_rescaled[n] = e_resc

        charges = list(e_rescaled.keys())
        energies = list(e_rescaled.values())

        ax = plt.gca()
        if not ax.lines:
            plt.hlines(0,xmin=0,xmax=1,linestyles='dashed',label='Exact',color='k')
        plt.plot(charges,energies,'o--',label=name)
        if new_figure:
            width = max([abs(e) for e in energies])
            plt.ylim(-3*width,+3*width)
        plt.xlabel('Fractional charge')
        plt.ylabel('Energy (eV)')
        if legend_out:
            plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        else:
            plt.legend()
        plt.grid()
            
        return plt
            
    
    def plot_U_dependency(self,element,lattice_parameter='a'):
        """
        Plot dependency of lattice parameter and energy gap on U parameter assigned
        to a specific element

        Parameters
        ----------
        element : (str)
            String of element to which U parameter is applied.
        lattice_parameter : (str), optional
            Lattice parameter whose dependency is evaluated. The default is 'a'.

        Returns
        -------
        plt :
            Matplotlib object
        """
        el = Element(element)
        plt.figure(figsize=(12,8))
        U_list = []
        lattice_params = []
        band_gaps = []
        for j in self.jobs:
            U_list.append(j.hubbard[el])
            lattice = j.final_structure.lattice
            lattice_params.append(getattr(lattice,lattice_parameter))
            band_gaps.append(j.energy_gap)
        
        plt.subplot(1,2,1)
        plt.xlabel('U on %s'%element)
        plt.ylabel(lattice_parameter+' ($\AA$)')
        plt.plot(U_list,lattice_params,'o--')
        plt.grid()
        
        plt.subplot(1,2,2)
        plt.xlabel('U on %s'%element)
        plt.ylabel('Energy gap (eV)')
        plt.plot(U_list,band_gaps,'o--')
        plt.grid()
        
        return plt
            
                
        
        
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 14 12:56:45 2020

@author: villa
"""

import os
import re
import warnings

def job_status(path=None,job_script_filename='job_vasp.sh'):
        
    if path == None:
        path = os.getcwd()
    os.system('squeue -o "%.10i %.9P %.40j %.8u %.2t %.10M %.5D %R" > job_list.txt')

    file = os.path.join(path,job_script_filename)
    if os.path.isfile(file):
        job_vasp = open(file)
       
    for line in job_vasp:
        # emulating 'grep' command
        job_line = re.findall(r'#SBATCH --job-name=', line)                    
        if job_line:
            job_name = line.split('=')[1].strip()
            
    
    with open('job_list.txt') as jobs_file:
        
        count = 0
        status_string = None        

        for line in jobs_file:
            split_line = line.split()
            name = split_line[2]
            status = split_line[4]            
 
            if name == job_name:
                
                if status == 'R':
                    status_string = 'RUNNING'
                    count += 1
                elif status == 'PD':
                    status_string = 'PENDING'

        if count > 1:
            warnings.warn(f': More than one job named "{job_name}" are simultaneously running\n',UserWarning)
        
        if not status_string:    
            status_string = 'NOT IN QUEUE'

    os.remove('job_list.txt') 
	        
    return status_string



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 26 13:59:39 2023

@author: villa
"""

import argparse
import os
import warnings

from pynter.cli.automations import setup_automation
from pynter.cli.analysis import setup_analysis
from pynter.cli.config import setup_config, run_config
from pynter.cli.defects import setup_defects
from pynter.cli.hpc import setup_hpc
from pynter.cli.inputs import setup_inputs
from pynter.cli.materials_project import setup_mp_database
from pynter.cli.plotter import setup_plotter
from pynter.cli.slurm import setup_job_script
from pynter.cli.phase_diagram import setup_phase_diagram


def main():
    """
    Handle main.
    """    
    parser = argparse.ArgumentParser(
        description="""
        This script works based on several sub-commands with their own options.
        To see the options for the sub-commands, type "pynter sub-command -h".
        """
        )

    subparsers = parser.add_subparsers()
    
    homedir = os.getenv("HOME")
    config_exists = os.path.exists(os.path.join(homedir,'.pynter/config.yml'))
    config_vasp_exists = os.path.exists(os.path.join(homedir,'.pynter/vasp.yml'))
    if config_exists is False and config_vasp_exists is False:
        warnings.warn('configuration files do not exist, run "pynter configure" in the terminal to create them')


    setup_analysis(subparsers)
    setup_automation(subparsers)  
    setup_config(subparsers)
    setup_defects(subparsers)
    setup_hpc(subparsers)
    setup_inputs(subparsers)
    setup_job_script(subparsers)
    setup_mp_database(subparsers)
    setup_phase_diagram(subparsers)
    setup_plotter(subparsers)
    
    args = parser.parse_args()
    
    try:
        args.func
    except AttributeError:
        parser.print_help()
        raise SystemExit("Please specify a command.")
    return args.func(args)

if __name__ == '__main__':
    main()